;;; gauche-mode.el ---  Major mode for gauche(gosh).

;; Keywords: Gauche, scheme, languages

;; This file base on "21.4 (patch 12) \"Portable Code\" XEmacs
;; Lucid & xemacs-sumo"'s scheme-mode.el apropos.el,
;; font-lock.el, lisp.el, help.el, hyper-apropos.el, texinfo.el and window.el
;;
;; Many code was copied from XEmacs.

;; Copyright (C) 1985, 1986, 1989, 1992-5, 97, 1998, 2000, 2001
;;					Free Software Foundation, Inc.
;; Copyright (C) 1995 Amdahl Corporation.
;; Copyright (C) 1996, 2000, 2001 Ben Wing.
;; Copyright (C) 2004-2007  NISHIO Yasuhiro
;;
;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License
;; as published by the Free Software Foundation; either version 2
;; of the License, or (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;;

;;; $Id: gauche-mode.el,v 2.66 2007/04/14 15:23:30 nishio Exp $

;; Add following code in your ~/.xemacs/init.el
;;
;;(setq auto-mode-alist
;;      (cons '("\\.scm$" . gauche-mode)
;;	      auto-mode-alist))
;;
;;(autoload 'gauche-mode "gauche-mode" "Major mode for gauche." t nil)
;;
;;(setq gauche-sym-file-name "File name for Modules & Symbols list")
;;      gauche-doc-file-name "File name for gauche documentation")
;;
;;If you use font-lock minor mode
;;(add-hook 'gauche-mode-hook
;;	  (lambda ()
;;	    (font-lock-mode 1)
;;	    (redraw-modeline)))
;;

(defconst gauche-mode-version-string
  "$Id: gauche-mode.el,v 2.66 2007/04/14 15:23:30 nishio Exp $")

(eval-when-compile
  (require 'custom)
  (require 'cl)

  (defmacro go-intern (str)
    `(intern ,str *gauche-obarray*))
  (defmacro go-intern-soft (str)
    `(intern-soft ,str *gauche-obarray*))
  (defmacro get-keyword (key lst)
    `(plist-get ,lst ,key))
  (defmacro string-join (lst sep)
    `(mapconcat #'identity ,lst ,sep))

  ;; use value cell for (class . module) list
  (defmacro go-props (sym)
    (let ((gsym (make-symbol "gsym")))
      `(let ((,gsym ,sym))
	 (and (boundp ,gsym) (symbol-value ,gsym)))))
  (defmacro go-set-props (sym props)
    `(set ,sym ,props))
  (defmacro go-add-prop (sym prop)
    (let ((gsym (make-symbol "gsym"))
	  (gprop (make-symbol "gprop")))
      `(let* ((,gsym ,sym)
	      (,gprop ,prop)
	      (props (go-props ,gsym)))
	 (or (member ,gprop props)
	     (set ,gsym (cons ,gprop props))))))
  ;;
  (defmacro go-set-module-prop (name)
    `(put ,name 'module t))
  (defmacro go-modulep (name)
    `(get ,name 'module))
  ;;
  (defmacro gauche-display-momentary-string (message face timeout)
    `(let ((g (make-glyph (vector 'string
				  :data
				  ,message)))
	   a)
       (unwind-protect
	   (progn
	     (setq a (make-annotation g (point) 'text nil nil nil t))
	     (set-glyph-face g ,face)
	     (or (input-pending-p)
		 (sit-for ,timeout)))
	 (delete-annotation a))))

  (defmacro with-progress-message (msg &rest body)
    (let ((gmsg (make-symbol "gmsg")))
      `(let ((,gmsg ,msg))
	 (lmessage 'progress "%s..." ,gmsg)
	 ,@body
	 (lmessage 'progress "%s...done." ,gmsg)))))

(require 'lisp-mode)
(require 'font-lock)
(require 'info-look)
(require 'annotations)		;for gauche documentation
(require 'regexp-opt)
;;(require 'timer-funcs)

(defgroup gauche-mode-faces nil
  "Editing Gauche code"
  :group 'faces)

;;; faces

;; faces for scheme codes
(defface gauche-procedure-face
  '((((class color) (background dark))  (:foreground "blue"))
    (((class color) (background light)) (:foreground "Blue4")))
  "procedure face for gauche-mode"
  :group 'gauche-mode-faces)

(defface gauche-generic-function-face
  '((((class color) (background dark))  (:foreground "blue"))
    (((class color) (background light)) (:foreground "Blue3")))
  "Ganeric function face for gauche-mode"
  :group 'gauche-mode-faces)

(defface gauche-syntax-face
  '((((class color) (background dark))  (:foreground "red"))
    (((class color) (background light)) (:foreground "Red4"))
    (t (:bold t)))
  "Syntax face for gauche-mode"
  :group 'gauche-mode-faces)

(defface gauche-class-name-face
  '((((class color) (background dark))  (:foreground "Cyan"))
    (((class color) (background light)) (:foreground "Cyan4"))
    (t (:bold t)))
  "Class Name face for gauche-mode"
  :group 'gauche-mode-faces)

(defface gauche-keyword-symbol-face
  '((((class color) (background dark))  (:foreground "Purple"))
    (((class color) (background light)) (:foreground "Purple3"))
    (t (:bold t)))
  "Keyword face for gauche-mode"
  :group 'gauche-mode-faces)

(defface gauche-variable-face
  '((((class color) (background dark))  (:foreground "Magenta"))
    (((class color) (background light)) (:foreground "Magenta4"))
    (t (:bold t)))
  "Variable face for gauche-mode."
  :group 'gauche-mode-faces)

;; face for Documentation
(defface gauche-doc-code-face
  '((((class color) (background dark))  (:foreground "blue"))
    (((class color) (background light)) (:foreground "blue4"))
    (t (:bold t)))
  "@Code face for gauche documentation."
  :group 'gauche-mode-faces)

(defface gauche-doc-var-face
  '((((class color) (background dark))  (:foreground "blue"))
    (((class color) (background light)) (:foreground "Magenta4" :italic t))
    (t (:bold t)))
  "@var face for gauche documentation."
  :group 'gauche-mode-faces)

(defface gauche-doc-example-face
  '((((class color) (background dark))  (:foreground "gray90"))
    (((class color) (background light)) (:foreground "gray20")))
  "@example face for gauche documentation."
  :group 'gauche-mode-faces)

;; xxx
(defface gauche-output-display-face
  '((((class color) (background dark))  (:foreground "green"))
    (((class color) (background light)) (:foreground "green4"
						     :background "gray80")))
  "gauche process output"
  :group 'gauche-mode-faces)

(defface gauche-output-values-face
  '((((class color) (background dark))  (:foreground "blue"))
    (((class color) (background light)) (:foreground "black"
						     :background "gray90")))
  "gauche process output"
  :group 'gauche-mode-faces)

(defface gauche-output-error-face
  '((((class color) (background dark))  (:foreground "red"))
    (((class color) (background light)) (:foreground "black" :bold t)))
  "gauche process output"
  :group 'gauche-mode-faces)
;;
(defgroup gauche nil
  "Editing Gauche code"
  :group 'scheme)

(defcustom gauche-obarray-size 1223
  "*Gauche obarry vector size"
  :group 'gauche
  :type  'integer)

(defcustom gauche-mode-hook nil
  "*Hook to run when entering gauche-mode.

for example:
	(lambda ()
	   (paren-set-mode 2)
	   (setq blink-matching-paren t)
	   (font-lock-mode 1)
	   (redraw-modeline)))"
  :group 'gauche
  :type  'hook)

(defcustom gauche-sym-file-name nil
  "*File name for Gauche symbol definition.
file format:
  ((\"module-name\" (\"<class-name>\" . \"symbol-name\")...)
   ...)"
  :group 'gauche
  :type  'string)

(defcustom gauche-doc-file-name nil
  "*File name for Gauche doc data file.
file format:
  ((((type \"symbol-name\" [. \"symbol-args\"])...) . \"doc-string\")
   ....)"
  :group 'gauche
  :type  'string)

(defcustom gauche-fontify-all-modules t
  "*If non-nil, all interned(*gauche-obarray*) symbols are fontified.
If nil, see ``gauche-modules-for-fontify''."
  :group 'gauche
  :type  'boolean)

(defcustom gauche-modules-for-fontify
  '(null scheme srfi-6 srfi-8 srfi-10 srfi-17 gauche gauche.gf gauche.ioutil
	 gauche.object user)
  "*module list to fontify symbols.
This variable is ignored if gauche-fontify-all-modules is t"
  :group 'gauche)

(defcustom gauche-indent-regexp-alist '(("^define-\\w"    . defun)
					("^call-with-\\w" . 1)
					("^with-\\w"      . 1))
  "*alist for syntax/procedure indentation.
for example;
  '((regexp-string . indentation-word)...)

To override this indentation style, you can use gauche-indent-prop-list
variable."
  :group 'gauche)

(defcustom gauche-indent-prop-list nil
  "*property list for syntax/procedure indentation.
for example;
  '(((gauche-indent-function 0) syntax1 syntax2...)
    ((gauche-indent-function 1) syntax1 syntax2...))

See also gauche-indent-regexp-alist"
  :group 'gauche)

(defcustom gauche-mode-keybind-list nil
  "*gauche-mode keybinding list. You can customize the gauche-mode
keybindings.
for example;
  '(([(control c) i]   interactive-function)
    (key-sequence      interactive-function)...)"
  :group 'gauche)

(defcustom gauche-auto-start-inferior-process
    ;; xxx
    ;; change var name [2005-11-17]
  (or (and (boundp 'gauche-autostart-inferior-process)
	   gauche-autostart-inferior-process)
      t)
  "*Non-nil meens auto start gauche process when entering gauche-mode."
  :group 'gauche
  :type 'boolean)
(make-obsolete-variable 'gauche-autostart-inferior-process
			'gauche-auto-start-inferior-process)
;;
(defcustom gauche-complete-symbol-string "[complete]"
  "*T.B.D."
  :group 'gauche
  :type 'string)
(defcustom gauche-complete-symbol-timeout 0.8
  "*T.B.D."
  :group 'gauche)
(defcustom gauche-nothing-to-complete-symbol-string "[no match]"
  "*T.B.D."
  :group 'gauche
  :type 'string)

;;XXX
(defcustom gauche-module-ignore-interning '()
  "*module list that dose not intern to *gauche-obarray*.
for example;
  '(gtk gtk\.error-dialog scmail scmail\.bayesian-filter scmail\.config
	scmail\.mail scmail\.progress scmail\.uti)"
  :group 'gauche)

(defgroup gauche-documentation nil
  "Gauche documentation"
  :group 'gauche)

(defcustom gauche-doc-left-margin 2
  "*left margin for gauche documentation."
  :group 'gauche-documentation
  :type  'integer)

(defcustom gauche-doc-right-margin 4
  "*right margin for gauche documentation."
  :group 'gauche-documentation
  :type  'integer)

(defcustom gauche-doc-itemize-left-margin 4
  "*left margin for gauche documentation."
  :group 'gauche-documentation
  :type  'integer)

(defcustom gauche-doc-table-left-margin 8
  "*left margin for gauche documentation."
  :group 'gauche-documentation
  :type  'integer)

(defcustom gauche-doc-example-left-margin 2
  "*left margin for gauche documentation."
  :group 'gauche-documentation
  :type  'integer)

;; gauche-mode vars.
(defvar *gauche-obarray* nil
  "array for Gauche auto-completion.")

(defvar *gauche-doc-db* nil
  "Gauche doc-db file handler. This handler was managed by gauche mode.
Do NOT set/modify this variable")

(defvar gauche-font-lock-keywords nil
  "font-lock-form list. See `font-lock-keywords' for general description.
On gauche mode, this variable will be initialized and override
by gauche-initialize-font-lock-keywords.")

;; for inferior process
(defcustom gauche-program-name "gosh"
  "*Program invoked by the gauche-start-inferior-repl command."
  :group 'gauche
  :type  'string)
(defcustom gauche-program-args '("-i" "-ligauche")
  "*Program arguments"
  :group 'gauche)

(defcustom gauche-message-timeout 60
  "*timeout for multi-line echo. Default is 60 seconds.
Wait specified seconds or until user input available."
  :group 'gauche
  :type 'number)

(defcustom gauche-function-args-idle-delay 0.5
  "*delay time for function args display. Default is half second."
  :group 'gauche
  :type 'number)

(defcustom gauche-auto-display-function-args t
  "*If non-nil, gauche-mode auto display function's args."
  :group 'gauche
  :type 'boolean)

(defvar gauche-function-args-idle-timer nil)

(defconst gauche-inferior-process-name "gauche process"
  "gauche process name")

(defconst gauche-inferior-output-buffer-name "*gauche output buffer*"
  "gauche output buffer name")

(defconst gauche-stream-name " *gauche-stream*")
(defvar *gauche-stream* (get-buffer-create gauche-stream-name))
(defvar *gauche-send-strings* nil)
(defvar gauche-process-filter-functions '(gauche-append-output))
(defvar gauche-process-filter-functions-on-last '(gauche-show-output))

(defvar gauche-process-ptyp process-connection-type
  "connection type of gauche process and xemacs.
On default we use process-connection-type. Some operating system has
problem on pty capabilities. If t, gauche-process use pty.
If you set this, you should check pty's owner and mode.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; gauche-mode

(define-derived-mode gauche-mode scheme-mode "Gauche"
  "Major mode for Gauche(scheme) code.
This mode is derived from scheme-mode. Editing commands are similar
to those of lisp-mode.

Commands:
Delete converts tabs to spaces as it moves back.
Blank lines separate paragraphs.  Semicolons start comments.
\\{gauche-mode-map}
Entry to this mode calls the value of gauche-mode-hook
if that value is non-nil.

Note: gauche-mode dose NOT enable font-lock minor mode.
You can use gauche-mode-hook to enable font-lock mode."

  (make-local-variable 'lisp-indent-function)
  (setq lisp-indent-function 'gauche-indent-function)
  ;;(setq scheme-mit-dialect nil)

  (unless *gauche-obarray*
    (gauche-initialize-obarray gauche-sym-file-name)
    (gauche-initialize-document-props)
    (gauche-initialize-keymap gauche-mode-map)
    (gauche-initialize-indentation))

  (unless gauche-font-lock-keywords
    (gauche-initialize-font-lock-keywords))

  (if (and gauche-auto-start-inferior-process
	   gauche-process-ptyp)
      (gauche-start-inferior-repl))

  (if gauche-auto-display-function-args
    (gauche-toggle-auto-display-function-args t)))

(defun gauche-initialize-indentation ()
  (dolist (lst (list '(((gauche-indent-function 0)
			begin
			begin0
			call-with-current-continuation
			call-with-output-string
			call/cc
			delay
			rxmatch-cond
			with-output-to-string)
		       ((gauche-indent-function 1)
			and-let*
			case
			dolist
			dotimes
			eval-when
			fluid-let
			hash-table-for-each
			hash-table-map
			lambda
			let*
			let*-values
			let-syntax
			let-values
			let/cc
			letrec
			letrec-syntax
			make
			parameterize
			parse-options
			rxmatch-case
			unless
			until
			when
			while)
		       ((gauche-indent-function 2)
			do
			let-args
			let-keywords*
			let-optionals*
			let1
			receive
			rxmatch-if
			rxmatch-let)
		       ((gauche-indent-function 3)
			dynamic-wind)
		       ((gauche-indent-function defun)
			define
			defmacro)
		       ((gauche-indent-function scheme-let-indent)
			let))
		     ;; User defined indentations
		     gauche-indent-prop-list))
    (dolist (alst lst)
      (dolist (sym (cdr alst))
	(apply #'put (go-intern (symbol-name sym)) (car alst))))))

(defun gauche-initialize-keymap (map)
  (dolist (lst (list
		'(([(control c) (control i)] gauche-complete-symbol)
		  ([(control c) (control d)] gauche-documentation-at-point)
		  ([(control c) (control f)] gauche-documentation-for-function)
		  ([(control c) (control l)] gauche-show-last-output)
		  ([(control c) (control m)] gauche-macroexpand-1-at-point)
		  ([(control c) return]      gauche-macroexpand-1-at-point)
		  ([(control c) (meta m)]    gauche-macroexpand-at-point)
		  ([(control c) (control r)] gauche-eval-region)
		  ([(control c) a]	     gauche-apropos)
		  ([(control c) f]	     info-lookup-symbol)
		  ([(control c) d]	     gauche-documentation)
		  ([(control c) m]	     gauche-which-module-at-point)
		  ([(control c) s]	     gauche-start-inferior-repl)
		  ([(control c) i]	     gauche-inferior-complete-symbol)
		  ([(control c) tab]	     gauche-complete-symbol)
		  ([(control h) d]	     gauche-inferior-describe-symbol)
		  ([(control h) a]	     gauche-inferior-apropos)
		  ([(control x) (control e)] gauche-eval-last-sexp)
		  ([(meta control g)]	     gauche-send-interrupt)
		  ([(meta control i)]	     gauche-inferior-complete-symbol)
		  ;;([(meta control i)]	     info-complete-symbol)
		  ([(meta control x)]	     gauche-eval-definition)
		  ([(meta j)]		     gauche-eval-last-sexp-and-print)
		  ([(meta tab)]		     gauche-inferior-complete-symbol))
		;;User defined keybindings
		gauche-mode-keybind-list))
    (dolist (elt lst)
      (apply #'define-key map elt))))

(defun gauche-mode-update-font-lock-mode (keywords case-fold)
  (put 'gauche-mode 'font-lock-defaults
       `(,keywords			; keywords
	 nil				; keywords-only
	 ,case-fold			; case-fold search
	 ((?: . "w") (?- . "w") (?* . "w") (?+ . "w") (?. . "w") (?< . "w")
	  (?> . "w") (?= . "w") (?! . "w") (?? . "w") (?$ . "w") (?% . "w")
	  (?_ . "w") (?& . "w") (?~ . "w") (?^ . "w") (?/ . "w"))
	 beginning-of-defun))
  (cond (font-lock-mode
         (setq font-lock-keywords nil)
         (font-lock-set-defaults t))))

;; util functions for completion.

(defun gauche-syntax-or-procedure-p (sym)
  "Return non-nil if SYM is syntax, macro or procedure.
SYM should be in *gauche-obarray*."
  (let ((props (go-props sym)))
    (or (assq (go-intern "<procedure>") props)
	(assq (go-intern "<syntax>")    props)
	(assq (go-intern "<generic>")   props)
	(assq (go-intern "<macro>")     props))))

(defun gauche-module-p (sym)
  "Return non-nil if SYM is module. SYM should be in *gauche-obarray*."
  (go-modulep sym))

(defun gauche-symbol-p (sym)
  "Return non-nil if SYM has value.
SYM should be in *gauche-obarray*."
  (go-props sym))

(defun gauche-complete-symbol ()
  "Perform completion on gauche symbol preceding point.
Compare that symbol against the known gauche symbols.

The context determines which symbols are considered. If the symbol
starts just after an open-parenthesis, only syntaxes and functions.
In use form, only modules.  Otherwise, all symbols with function
definitions, values or properties are considered."
  (interactive)
  (gauche-complete-symbol-1 :function    (list *gauche-obarray*
					       #'gauche-syntax-or-procedure-p)
			    :keyword     (list #'gauche-keyword-completions)
			    :module      (list *gauche-obarray*
					       #'gauche-module-p)
			    :all-modules (list *gauche-obarray*
					       #'gauche-module-p)
			    :default     (list *gauche-obarray*
					       #'gauche-symbol-p)))

;;;;; from xemacs-21.4.12/lisp/lisp.el
(defun gauche-complete-symbol-1 (&rest coll-pred-list)
  (let* ((end        (save-excursion
		       (skip-chars-forward "^ \t\n()[]{}|#'\"")
		       (point)))
	 (beg        (with-syntax-table gauche-mode-syntax-table
		       (let* ((c  (char-before)) ;nil if (= 1 (point))
			      (cc (and c (char-syntax c))))
			 (if (or (eq cc ?w) (eq cc ?_))	;w word, _ symbol-co
			     (save-excursion
			       (backward-sexp 1)
			       (while (eq (char-syntax (char-after)) ?\ ) ; ws
				 (forward-char 1))
			       (point))
			   (point)))))	; for null("")
	 (pattern    (buffer-substring-no-properties beg end))
	 (coll-pred  (if (eq (char-before beg) ?\()
			 (get-keyword :function coll-pred-list)
		       (let ((func (gauche-function-at-point))
			     (kcol (get-keyword :keyword coll-pred-list)))
			 (if (and kcol (eq (char-after beg) ?:))
			     (append kcol (list func))
			   (get-keyword (cond
					 ((member func '("extend"
							 "import"
							 "use"))
					  :module)
					 ((member func '("select-module"
							 "with-module"))
					  :all-modules)
					 (t
					  :default))
					coll-pred-list)))))
	 (completion (apply #'try-completion pattern coll-pred))
	 (tmpbuf     "*Completions*"))
    (cond ((eq completion t)
	   (let ((w (get-buffer-window tmpbuf)))
	     (if w
		 (delete-window w)
	       (gauche-display-momentary-string gauche-complete-symbol-string
						'bold
						gauche-complete-symbol-timeout)))
	   (lmessage 'no-log "complete."))
	  ((null completion)
	   (lmessage 'error "Can't find completion for \"%s\"" pattern)
	   (gauche-display-momentary-string gauche-nothing-to-complete-symbol-string
					    'red
					    gauche-complete-symbol-timeout)
	   (ding t 'no-completion))
	  ((not (string-equal pattern completion))
	   (delete-region beg end)
	   (insert completion))
	  (t
	   (if (not (= (point) end))
	       (goto-char end))
	   (with-progress-message "Making completion list"
	     ;; use temp-buffer-shrink-to-fit ?
	     ;; temp-buffer-max-height ?
	     (with-output-to-temp-buffer tmpbuf
	       (display-completion-list
		(sort (apply #'all-completions pattern coll-pred)
		      #'string-lessp))))))))

;;; util function
(defun gauche-interactive-argument (at-point-func prompt comps predicate
						  &optional require-match
						  init-contents history)
  (let ((sname (funcall at-point-func)))
    (list (let ((enable-recursive-minibuffers t))
	    (completing-read (concat prompt
				     (if (not (string-equal sname ""))
					 (format " (default %s)" sname))
				     ": ")
			     comps
			     predicate
			     require-match init-contents history sname)))))

;;current-word symbol-near-point
(defalias 'gauche-symbol-name-at-point 'current-word)

(eval-when-compile
  (defmacro go-function-at-point ()
    '(let ((obj (read (current-buffer))))
       (and (atom obj)
	    (gauche-syntax-or-procedure-p (go-intern-soft (symbol-name obj)))
	    (symbol-name obj)))))
;; from lisp.el
(defun gauche-function-at-point ()
  "Return the function whose name is around point.
If that gives no function, return the function which is called by the
list containing point.  If that doesn't give a function, return nil."
  (or (ignore-errors
	(with-syntax-table gauche-mode-syntax-table
	  (save-excursion
	    (or (not (zerop (skip-syntax-backward "_w")))
		(eq (char-syntax (char-after (point))) ?w)
		(eq (char-syntax (char-after (point))) ?_)
		(forward-sexp -1))
	    (skip-chars-forward "`'")
	    (go-function-at-point))))
      (ignore-errors
	(save-excursion
	  (save-restriction
	    (narrow-to-region (max (point-min) (- (point) 1000))
			      (point-max))
	    (backward-up-list 1)
	    (forward-char 1)
	    (go-function-at-point))))))

;; from: xemacs-21.4.12/lisp/help.el
(defun gauche-which-module-at-point (&optional arg)
  "Show module name that contains the symbol."
  (interactive "P")
  (if arg
      (call-interactively #'gauche-which-module)
    (gauche-which-module-1 (gauche-symbol-name-at-point))))

(defun gauche-which-module-1 (sname)
  (let ((sym (go-intern-soft sname)))
    (if (and sym (go-props sym))
	(message "`%s' is defined in `%s'." sname
		 (mapconcat (lambda (p)
			      (symbol-name (cdr p)))
			    (go-props sym)
			    "', `"))
      (message "void symbol %s ??" sname))))

(defun gauche-which-module (symname)
  "Show module name that contains the symbol.
When run interactively, it defaults to any symbol found by
`gauche-symbol-name-at-point'."
  (interactive (gauche-interactive-argument #'gauche-symbol-name-at-point
					    "Gauche Symbol"
					    *gauche-obarray*
					    #'gauche-symbol-p
					    t))
  (gauche-which-module-1 symname))

;;;
(defun gauche-initialize-obarray (sym-file)
  "Init gauche obarray."
  (cond ((and (stringp sym-file)
	      (file-exists-p sym-file))
	 (with-progress-message "Making Gauche symbol obarray"
	   (setq *gauche-obarray* (make-vector gauche-obarray-size 0))
	   (dolist (elm (with-temp-buffer
			  (insert-file-contents sym-file)
			  (read (current-buffer))))
	     (let ((mod (go-intern (car elm))))
	       (unless (memq mod gauche-module-ignore-interning)
		 (go-set-module-prop mod)
		 (dolist (pair (cdr elm))
		   (go-add-prop (go-intern (cdr pair))
				(cons (go-intern (car pair)) mod))))))))
	(t
	 (lmessage 'error "Gauche Symbol file ``%s'' does not exist!" sym-file)
	 (ding))))

(defun gauche-initialize-document-props ()
  ;; put doc props
  (map-database (lambda (k v)
		  (if (not (string-match "^doc\0" k))
		      (put (go-intern k) :doc t)))
		(gauche-doc-db-handle)))

(eval-when-compile
  (defmacro gauche-mods-for-fontify ()
    '(mapcar (lambda (mod)
	       (go-intern-soft (symbol-name mod)))
	     gauche-modules-for-fontify))

  (defmacro gauche-classp (cstr)
    `(lambda (s)
       (assq (go-intern ,cstr)
	     (if gauche-fontify-all-modules
		 (go-props (go-intern-soft s))
	       (mapcar (lambda (mod)
			 (rassq mod (go-props (go-intern-soft s))))
		       (gauche-mods-for-fontify))))))

  (defmacro gauche-class-of-p (cslst)
    `(lambda (s)
       (do ((props (go-props (go-intern-soft s)))
	    (cslst ,cslst (cdr cslst)))
	   ((or (null cslst)
		(assq (go-intern-soft (car cslst))
		      (if gauche-fontify-all-modules
			  props
			(mapcar (lambda (mod)
				  (rassq mod props))
				(gauche-mods-for-fontify)))))
	    cslst)))))

(defun gauche-initialize-font-lock-keywords ()
  "Init font-lock-lock-keywords for gauche-mode.
This function initializes gauche-font-lock-keywords."
  (interactive)
  (setq gauche-font-lock-keywords
	(nconc (mapcar (lambda (lst)
			 (apply #'gauche-make-font-lock-form lst))
		       (list
			(list "(\\(\\w+\\)\\>"
			      (gauche-class-of-p '("<syntax>" "<macro>"))
			      '(1 gauche-syntax-face))
			(list "\\<\\w+\\>"
			      (gauche-classp "<generic>")
			      '(0 gauche-generic-function-face))
			(list "\\<\\w+\\>"
			      (gauche-classp "<procedure>")
			      '(0 gauche-procedure-face))
			(list "\\<\\w+\\>"
			      (gauche-class-of-p '("<integer>" "<real>"))
			      '(0 font-lock-constant-face))
			(list "\\<[^<:]\\w+\\>"
			      (lambda (s)
				(let ((sym (go-intern-soft s)))
				  (if (gauche-symbol-p sym)
				      (or gauche-fontify-all-modules
					  (do ((mods (gauche-mods-for-fontify)
						     (cdr mods))
					       (p (go-props sym)))
					      ((or (null mods)
						   (rassq (car mods) p))
					       mods))))))
			      '(0 gauche-variable-face))))
	       ;;
	       '(("\\<<\\sw+>\\>" . gauche-class-name-face)
		 ("\\<:\\sw+\\>" . gauche-keyword-symbol-face)
		 ("#/\\(.*?[^\\]\\)/i?" (1 font-lock-constant-face))
		 ;; from Alex Shinn (ML:gauche-devel@lists.sourceforge.net)
		 ;;'("#/\\(\\\\.\\|[^\\]\\)*/i?" . font-lock-string-face)
		 ;;'("#\\[\\([^]]*\\]\\|[^]]\\)*\\]" . font-lock-string-face)
		 )))
  (gauche-mode-update-font-lock-mode 'gauche-font-lock-keywords nil))

(defun gauche-make-font-lock-form (regexp predicate highlight)
  "Make font-lock-form for font-lock-keywords.

example:
       (gauche-make-font-lock-form \"(\\\\(\\\\w+\\\\)\\\\>\"
				    (lambda (str)
					(if (string-equal str \"abc\")
					    t
					  nil))
				    '(1 red)))
	=> (#<compiled function> (1 red))

	- search ``(WORD''
	- iff WORD equal \"abc\" then highlight ``abc''

	Car of HIGHLIGHT is used for determine the argument of PREDICATE."
  (list
   (byte-compile			;eval for debug
    `(lambda (limit)
       (let ((n ,(car highlight)))
	 (catch nil
	   (while (re-search-forward ,regexp limit t)
	     (if (funcall ,predicate
			  (buffer-substring-no-properties (match-beginning n)
							  (match-end n)))
		 (throw nil t)))))))
   highlight))

;; from: hyper-apropos.el, apropos.el
(defun gauche-apropos (regexp)
  "Display lists of symbols matching REGEXP."
  (interactive (list (read-string "Gauche Apropos(regexp): ")))
  ;; accum --- (("string1" . ((<class> . module) (<class> . module)))
  ;;		("string2" . ((<class> . module) (<class> . module)))
  ;;		("string3" . t)) ; for module
  (let (accum)
    (mapatoms (lambda (sym)
		(if sym
		    (let ((str (symbol-name sym)))
		      (if (string-match regexp str)
			  (push (cons str (or (go-props sym)
					      (go-modulep sym)))
				accum)))))
	      *gauche-obarray*)
    (if (not accum)
	(message "Nothing for `%s'." regexp)
      (with-displaying-help-buffer
       (lambda ()
	 (with-current-buffer standard-output
	   (insert-face (format "Apropos search for: %s\n\n" regexp) 'bold)
	   (let ((t-face-alist
		  (mapcar (lambda (lst)
			    (cons (go-intern (car lst)) (cdr lst)))
			  '(("<procedure>" ?p gauche-procedure-face)
			    ("<generic>"   ?g gauche-generic-function-face)
			    ("<syntax>"    ?S gauche-syntax-face)
			    ("<macro>"     ?M gauche-syntax-face)
			    ("<symbol>"    ?s font-lock-constant-face)
			    ("<integer>"   ?c font-lock-constant-face)
			    ("<real>"	   ?c font-lock-constant-face)
			    ("<class>"     ?\  gauche-class-name-face)))))
	     ;; sort gauche symbols
	     ;; plst --- ("string" . ((<class> . module) (<class> . module)))
	     (dolist (plst (sort accum
				 (lambda (&rest ab)
				   (apply #'string-lessp
					  (mapcar (lambda (a)
						    (downcase (car a))) ab)))))
	       ;;sort modules
	       ;; prop --- ((<class> . module) (<class> . module)) or t
	       (dolist (prop (cond
			      ((consp (cdr plst))
			       (sort (copy-list (cdr plst)) ; el sort == sort!
				     (lambda (&rest ab)
				       (apply #'string-lessp
					      (mapcar (lambda (a)
							(symbol-name (cdr a)))
						      ab)))))
			      ((eq (cdr plst) t)
			       '("<module>"))))
		 (apply (lambda (s m c f)
			  (insert (char-to-string c) "  ")
			  (let ((width (/ (window-width) 2))
				(e     (insert-face s f)))
			    (insert (make-string (max (- width (length s)) 2)
						 ?\ ))
			    (set-extent-property e 'mouse-face 'highlight))
			  (insert (format "\t%s\n" m)))
			;; args
			(car plst)	; s
			(if (consp prop) ; m
			    (concat "(" (symbol-name (cdr prop)) ")")
			  prop)		; "<module>"
			(cond ((stringp prop)
			       '(?m font-lock-constant-face))
			      ((cdr (assq (car prop) t-face-alist))) ; (c f)
			      ;; for <OBJECT-meta>
			      ((string-match "^<.*>$" (car plst))
			       '(?\  gauche-class-name-face))
			      (t
			       '(?\  gauche-variable-face))))))))))
      (format "Apropos for `%s'" regexp)))) ; buffer name

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; document

(eval-when-compile
  (defmacro gauche-doc-database (entry)
    `(read (get-database ,entry (gauche-doc-db-handle) "nil"))))

(defalias 'gauche-prepend-text-property 'font-lock-prepend-text-property)
(defalias 'gauche-append-text-property 'font-lock-append-text-property)

(define-key help-mode-map [button2] 'gauche-documentation-by-mouse-click)

(defun gauche-documentation-by-mouse-click (event)
  (interactive "e")
  (mouse-set-point event)
  (gauche-documentation (extent-string (extent-at (point) nil 'highlight))))

(defun gauche-documentation-at-point (&optional arg)
  "Show symbol's documentation."
  (interactive "P")
  (let ((sname (gauche-symbol-name-at-point)))
    (if (or arg (string-equal sname ""))
	(call-interactively #'gauche-documentation)
      (gauche-documentation sname))))

(defun gauche-documentation-for-function (func)
  (interactive
   (gauche-interactive-argument #'gauche-function-at-point
				"Gauche Function"
				*gauche-obarray*
				#'gauche-syntax-or-procedure-p
				t))
  (gauche-documentation func))

(defun gauche-documentation (symname)
  "Show documentation. When run interactively, it defaults to any symbol
found by `gauche-symbol-name-at-point'."
  (interactive
   (gauche-interactive-argument #'gauche-symbol-name-at-point
				"Gauche Doc"
				*gauche-obarray*
				(lambda (s)
				  (get s :doc))
				t))
  (if (string-equal symname "")
      (lmessage 'prompt "Please enter symbol name.")
    (let ((entry-list (gauche-doc-database symname)))
      (if entry-list
	  (gauche-documentation-1 entry-list symname)
	(message "No document available for `%s'." symname)))))

;;;

(eval-when-compile
  (defmacro assq-val (key alist)
    `(cdr (assq ,key ,alist)))

  ;;utils for gauche-texi-attrs
  (defmacro gd-texi-attr (op key)
    `(plist-get (gethash ,op gauche-texi-attrs) ,key))

  ;;ff util
  (defmacro gd-texi-cmd-str (&optional str face)
    (let ((gstr (make-symbol "gstr"))
	  (gface (make-symbol "gface")))
      `(let ((,gstr ,str)
	     (,gface ,face))
	(byte-compile `(lambda (op &rest rest)
			,(if ,gface
			     `(insert-face ,(or ,gstr
						`(symbol-name op)) ',,gface)
			     `(insert ,(or ,gstr
					   `(symbol-name op)))))))))
  (defmacro gd-texi-make-cmd-str-entry (ss &optional str face)
    (let ((gss (make-symbol "gss"))
	  (gstr (make-symbol "gstr")))
      `(let ((,gss ,ss)
	     (,gstr ,str))
	(list (intern ,gss)
	 :ff `,(gd-texi-cmd-str (or ,gstr ,gss) ,face)))))

  (defmacro gd-texi-cmd-face (face)
    `(lambda (op &rest rest)
       (if (looking-at "{")
	   (let ((e (save-excursion
		      (forward-sexp 1)
		      (delete-char -1)	; delete `}';
		      (point))))
	     (gauche-prepend-text-property (1+ (point)) e 'face ,face)
	     (delete-char 1))		; delete `{'
	 ;;(error "gauche-texi-cmd-face(%S)" op) ;xx
	 )))

  ;; check: receive,
  (defmacro gd-texi-cmd-ref (&optional xref)
    `(lambda (op &rest rest)
       (if (looking-at "{")
	   (let ((e (save-excursion
		      (forward-sexp 1)
		      (delete-char -1)	; delete ?}
		      (point))))
	     ,(if xref
		  ;; for 'xref
		  `(progn
		     (delete-char 1)	; delete `{'
		     (insert "See "))
		`(let ((ent (nth 2 (split-string (buffer-substring-no-properties
						  (1+ (point)) e) ","))))
		   (if ent
		       (progn
			 (delete-region (point) e)
			 (insert ent))
		     (delete-char 1))))) ; delete `}'
	 (error "gauche-texi-cmd-ref(%S)" op))))

  (defmacro gd-texi-cmd-wsp ()
    `(lambda (op &rest rest)
       (let ((s (symbol-name op)))
	 (put-text-property 0 1 'hard 't s)
	 (insert s))))

  (defmacro gd-texi-cmd-comment ()
    `(lambda (op &rest rest)
       (delete-region (point)
		      (let ((e (point-at-eol)))
			(+ e (if (eq (char-after e) ?\n)
				 1
			       0))))))
  (defmacro gd-texi-cmd-example ()
    `(lambda (op &rest rest)
       (and (looking-at "\n") (delete-char 1)))) ;delete-char returns nil

  (defmacro gd-texi-cmd-defvar (&optional defivar)
    `(lambda (op &rest rest)
       ;; defivar, new-env and @defivar->@devivarx
       ,(if defivar
	    `(save-excursion
	       (insert "@defivarx")))
       (list :style "+ ")))
  ;;
  (defmacro gd-env-get (env key)
    `(plist-get (cdr ,env) ,key))

  (defmacro gd-label-style (pos &optional buf)
    `(plist-get (cdr (extent-property (extent-at ,pos ,buf 'gauche-texi-env)
				      'gauche-texi-env))
		:style))

  (defmacro gd-make-left-margin (glyph pos)
    `(make-annotation ,glyph ,pos 'text))

  (defmacro gd-make-item-label (glyph pos)
    `(make-annotation ,glyph ,pos)))

(defvar gauche-texi-attrs
  (let ((h (make-hash-table :test #'eq :size 101 :rehash-size 1.5))
	(attrs `((,(intern " ")		;SPACE
		  :ff ,(gd-texi-cmd-wsp))
		 (,(intern "\n")	;NL
		  :ff ,(gd-texi-cmd-wsp))
		 (,(intern "\t")	;TAB
		  :ff ,(gd-texi-cmd-wsp))
		 (*
		  :ff ,(lambda (op ti-env)
			 (if (looking-at "[ \t]*$")
			     (delete-region (match-beginning 0)
					    (1+ (match-end 0)))) ;xxx
			 (let ((s "\n"))
			   (put-text-property 0 1 'hard 't s)
			   (insert s))))
		 ,(gd-texi-make-cmd-str-entry "?")
		 ;; check: ssax:make-elem-parser
		 ,(gd-texi-make-cmd-str-entry "@")
		 ,(gd-texi-make-cmd-str-entry "-")
		 ,(gd-texi-make-cmd-str-entry ".")
		 (:			; do nothing
		  :ff (lambda (&rest rest)))
		 (b
		  :ff ,(gd-texi-cmd-face 'bold))
		 ,(gd-texi-make-cmd-str-entry "bullet{}" "*")
		 (c
		  :ff ,(gd-texi-cmd-comment))
		 (cartouche)		; display with f-box
		 (clear)
		 (code
		  :ff ,(gd-texi-cmd-face 'gauche-doc-code-face))
		 (comment
		  :ff ,(gd-texi-cmd-comment))
		 (defivar
		   :ff ,(gd-texi-cmd-defvar 'defivar)
		   :fill t
		   :indent 2
		   :item t		;defivarx
		   :type new-env)
		 (defivarx
		   :ff gauche-texi-cmd-mark
		   :type mark)
		 (defvar		; gdbm-setopt
		   :ff ,(gd-texi-cmd-defvar)
		   :fill t
		   :indent 0
		   :type new-env)
		 (display
		  :ff ,(gd-texi-cmd-example)
		  :indent ,(+ gauche-doc-example-left-margin 2)
		  :type new-env)	;indented disp
		 (dmn)
		 ,(gd-texi-make-cmd-str-entry "dots{}" "...")
		 (emph
		  :ff ,(gd-texi-cmd-face 'bold-italic))
		 (end
		  :ff gauche-texi-cmd-block
		  :type exit-env)
		 (enumerate
		  :ff gauche-texi-cmd-block
		  :type new-env
		  :fill t
		  :count-item t
		  :item t
		  :indent ,gauche-doc-itemize-left-margin)
		 ,(gd-texi-make-cmd-str-entry "equiv{}" " == ")
		 ,(gd-texi-make-cmd-str-entry "error" " error--> ")
		 (example
		  :ff ,(gd-texi-cmd-example)
		  :indent ,gauche-doc-example-left-margin
		  :type new-env)
		 ,(gd-texi-make-cmd-str-entry "expansion" " ==> ")
		 (file
		  :ff ,(gd-texi-cmd-face 'gauche-doc-code-face))
		 (flushleft
		  :type new-env)
		 (flushright
		  :type new-env)
		 (footnote)
		 (format
		  :ff ,(gd-texi-cmd-example)
		  :indent 0
		  :type new-env)	;no-indent disp
		 (ftable
		  :ff gauche-texi-cmd-block
		  :fill t
		  :item t
		  :type new-env)
		 (group
		  :ff gauche-texi-cmd-block
		  :type new-env)
		 (i
		  :ff ,(gd-texi-cmd-face 'italic))
		 (ifset
		  :type new-env)
		 (ifclear
		  :type new-env)
		 (item
		  :ff gauche-texi-cmd-mark
		  :fill t
		  :type mark)
		 (itemx
		  :ff gauche-texi-cmd-mark
		  :fill t
		  :type mark)
		 (itemize
		  :ff gauche-texi-cmd-block
		  :fill t
		  :indent ,gauche-doc-itemize-left-margin
		  :item t
		  :type new-env)
		 (kbd)
		 (key)
		 (lisp
		  :ff ,(gd-texi-cmd-example)
		  :indent ,gauche-doc-example-left-margin
		  :type new-env)
		 (math)
;		 (minus
;		  :ff ,(gd-texi-cmd-str "-"))
		 ,(gd-texi-make-cmd-str-entry "minus" "-")
		 (multitable
		  :ff gauche-texi-cmd-block
		  :item t
		  :type new-env
		  :indent 8)
		 (need)
		 (noindent)
		 (page
		  :ff (lambda (&rest rest)))
		 ,(gd-texi-make-cmd-str-entry "point" "-!-")
		 ,(gd-texi-make-cmd-str-entry "print" " -| ")
		 (quotation
		  :ff gauche-texi-cmd-block
		  :indent 6
		  :fill t
		  :type new-env)
		 (r
		  :ff ,(gd-texi-cmd-face 'default))
		 (ref
		  :ff ,(gd-texi-cmd-ref))
		 ,(gd-texi-make-cmd-str-entry "result{}" " => ")
		 (samp
		  :ff ,(lambda (op &rest rest)
			 (cond
			  ((looking-at "{")
			   (save-excursion
			     (forward-sexp 1)
			     (delete-char -1) ;delete `}'
			     (insert ?'))
			   (delete-char 1)
			   (insert ?`))
			  (t
			   (error "oops: @samp")))))
		 (set)
		 (smallexample
		  :ff ,(gd-texi-cmd-example)
		  :type new-env)
		 (smalllisp
		  :ff ,(gd-texi-cmd-example)
		  :type new-env)
		 (sp)
		 (strong)
		 (tab
		  :ff gauche-texi-cmd-mark
		  :type mark)
		 (table
		  :ff gauche-texi-cmd-block
		  :fill t
		  :indent ,gauche-doc-table-left-margin
		  :first-row-width 8
		  :item t
		  :type new-env)
		 (tex			;TeX command
		  :ff gauche-texi-cmd-block
		  :type new-env)
		 (uref)
		 (url)			;xxx
		 (value)		;@value{symbol}
		 (var
		  :ff ,(gd-texi-cmd-face 'gauche-doc-var-face))
		 (vtable
		  :ff gauche-texi-cmd-block
		  :item t
		  :fill t
		  :type new-env)
		 (w)			;cannot break word
		 (xref			; @xref{string}
		  :ff ,(gd-texi-cmd-ref 'xref))
		 ,(gd-texi-make-cmd-str-entry "{")
		 ,(gd-texi-make-cmd-str-entry "}")
		 (default
		   :type new-env
		   :fill t
		   :indent ,gauche-doc-left-margin))))
    (dolist (attr attrs h)
      (puthash (car attr) (cdr attr) h))))

(defvar gauche-texi-cmd-regexp  nil
  "T.B.D.")

(defun gauche-texi-cmd-regexp (attrs)
  (or gauche-texi-cmd-regexp
      (setq gauche-texi-cmd-regexp (let (specs words)
				     (maphash (lambda (sym val)
						(let ((s (symbol-name sym)))
						  (if (string-match "^\\w+$" s)
						      (push s words)
						    (push s specs))))
					      attrs)
				     (concat "@\\(\\(?:"
					     (regexp-opt specs)
					     "\\)\\|\\(?:"
					     (regexp-opt words)
					     "\\)\\>\\)?")))))
;; from texinfo.el
(defvar gauche-texi-syntax-table
  (let ((table (make-syntax-table)))
    (dolist (elm
	     '((?\"  ".")
	       (?\\  ".")
	       (?@   "_")
	       (?\^q "\\")
	       ;;(?\[  "(]")
	       ;;(?\]  ")[")
	       (?\[  "_")
	       (?\]  "_")
	       (?<   "_")
	       (?>   "_")
	       (?{   "(}")
	       (?}   "){")
	       (?\n  ">")
	       (?\'  "w"))
	     table)
      (apply #'modify-syntax-entry
	     (append elm (list table))))))

(defun gauche-doc-detexi-and-set-attr (beg end &optional ti-env)
  "deTexinfo command and set its attribute(texi cmd convert to extent attr)
TI-ENV is TexInfo envronment list, It is '(NAME :attr1 val1 :attr2 val2 ...)"
  (let ((texicmd (gauche-texi-cmd-regexp gauche-texi-attrs))
	(left-margin (and ti-env
			  (+ left-margin (gd-texi-attr (car ti-env) :indent))))
	(marks))
    (save-restriction
      (narrow-to-region beg end)
      (goto-char beg)
      (catch 'return
	(while (re-search-forward texicmd nil t)
	  (if (not (match-string 1))	;texicmd --- @\\(cmd\\)?
	      ;; unknown command
	      (save-excursion
		(looking-at "\\w+")
		(gauche-prepend-text-property (match-beginning 0)
					      (match-end 0)
					      'face
					      'red))
	    ;;;;;;;;;;;;;;;;;;;;
	    ;;process @cmd
	    (let* ((op (intern (match-string 1)))
		   (ff (or (gd-texi-attr op :ff)
			   ;;function for unknown @cmd
			   (lambda (op ti-env)
			     (let ((s (format "@%s" op)))
			       (gauche-prepend-text-property 0 (length s)
							     'face 'green s)
			       (insert s))))))
	      ;;delete "@cmd"
	      (delete-region (match-beginning 0) (match-end 0))
	      ;; dispatch
	      (let ((ret (funcall ff op ti-env)))
		(case (gd-texi-attr op :type)
		  ((exit-env)
		   ;; make extent for env
		   (set-extent-property (make-extent (point-min) (point))
					'gauche-texi-env
					(cons (car ti-env)
					      (plist-put (cdr ti-env)
							 :indent left-margin)))
		   ;; make extent for pars
		   (when marks
		     (cond
		      ((eq (car ti-env) 'multitable)
		       (let ((cws (reverse (gd-env-get ti-env :colwidths))))
			 (do ((p marks (cdr p))
			      (e (point) (cdar p))
			      (ws cws (if (eq (caar p) 'item)
					  cws
					(cdr ws))))
			     ((null p)))))
		      (t
		       (let ((style (gd-env-get ti-env :style))
			     (start (gd-env-get ti-env :start))
			     (flface (gd-env-get ti-env :first-line-face)))
			 (do ((p marks (cdr p))
			      (e (point) (cdar p))
			      (n (if (gd-texi-attr (car ti-env) :count-item)
				     (1- (apply #'+ (mapcar
						     (lambda (elm)
						       (if (eq (car elm) 'item)
							   1
							 0))
						     marks))))
				 (and n (if (eq (caar p) 'item)
					    (1- n)
					  n))))
			     ((null p))
			   (cond
			    ;; @table @itam par(first-row) cont-par(second-row)
			    (flface
			     (save-excursion
			       (goto-char (cdar p))
			       (skip-chars-forward "^\n") ;(end-of-line)?
			       (let ((ext (make-extent (cdar p) (1+ (point))))
				     (frw (gd-texi-attr (car ti-env)
							:first-row-width)))
				 (set-extent-property ext
						      'gauche-texi-env
						      (list (caar p) :indent (- left-margin frw)))
				 (set-extent-property (make-extent (1+ (point))
								   e)
						      'gauche-texi-env
						      (list 'item :indent left-margin)))))
			    ;;@defivarx
			    ((and style (eq (caar p) 'defivarx))
			     (save-excursion
			       (goto-char (cdar p))
			       (end-of-line)
			       (set-extent-property (make-extent (cdar p)
								 (1+ (point)))
						    'gauche-texi-env
						    (list (caar p)
							  :indent left-margin))
			       (let ((g (make-glyph (vector 'string :data style))))
				 (set-glyph-face g 'default)
				 (gd-make-item-label g (cdar p)))))
			    ;; @itemize/@enumerate @item
			    ((and style (eq (caar p) 'item))
			     (set-extent-property (make-extent (cdar p) e)
						  'gauche-texi-env
						  (list (caar p)
							:indent left-margin))
			     (let ((g (make-glyph
				       (vector 'string
					       :data
					       (apply #'format
						      style
						      (if n
							  (list (+ start n))
							nil))))))
			       (set-glyph-face g 'default)
			       (gd-make-item-label g (cdar p))))))))))
		   (throw 'return op))
		  ;;@enumarate/itemize/table ...
		  ((new-env)
		   (gauche-doc-detexi-and-set-attr (point) (point-max)
						   (cons op ret)))
		  ;;@item/tab
		  ((mark)
		   (push (cons op ret) marks)))))))))))

(defun gauche-read-from-string (string &optional noerror)
  (with-temp-buffer
    (insert string)
    (goto-char (point-min))
    (let ((buf (current-buffer))
	  acc pos)
      (condition-case err
	  (while t
	    (skip-chars-forward "[ \t\n]")
	    (setq pos (point)
		  acc (cons (read buf) acc)))
	(end-of-file (cons (reverse acc)
			   (if (= pos (point-max))
			       nil
			     (buffer-substring-no-properties pos
							     (point-max)))))
	(t (if noerror
	       (cons (reverse acc)
		     (buffer-substring-no-properties pos (point-max)))
	     (apply #'error err)))))))

;; @item, @itemx, @tab, @defivar
(defun gauche-texi-cmd-mark (op ti-env)
  (let ((b (point)))
    (skip-chars-forward " \t\n")
    (delete-region b (point)))
  (case op
    ((item itemx)
     (let ((face (gd-env-get ti-env :first-line-face)))
       (when face
	 (let ((pos (point)))
	   (save-excursion
	     (skip-chars-forward "^\n")
	     (gauche-prepend-text-property pos (point) 'face face)
	     (gauche-prepend-text-property pos (point) 'face 'bold))))))
    ;;XXX
    ((tab)
     ;;
     )
    ((defivarx)
     (when (looking-at "\\s-*{")	; @defivarx {<class> var}
       (save-excursion
	 (forward-sexp 1)
	 (delete-char -1))
       (delete-region (match-beginning 0) (match-end 0)))
     (let ((pos (point)))
       (save-excursion
	 (skip-chars-forward "^\n")
	 ;;xxx sbr
	 (gauche-prepend-text-property pos (point) 'face 'bold))))
    (t
     (error "oops: unknown op(%S)" op)))
  (point))

;; xxx macro or sep func
(defun gauche-texi-cmd-block (op ti-env)
  (case op
    ((end)
     (let* ((e (point-at-eol))
	    (ename (let ((elm (gauche-read-from-string
			       (buffer-substring-no-properties (point) e))))
		     (and (null (cdr elm))
			  (eq (caar elm) (car ti-env))
			  (caar elm)))))

       (if (not ename)
	   (error "end oops:@%s %s/%s"
		  op
		  (buffer-substring-no-properties (point) e)
		  (car ti-env)))
       (delete-region (point) e)
       (and (looking-at "\n") (delete-char 1))
       ename))
    ((itemize)
     (let* ((e     (point-at-eol))
	    (elm   (gauche-read-from-string
		    (buffer-substring-no-properties (point) e)))
	    (style (and (null (cdr elm))
			(>= 1 (length (car elm)))
			(assq-val (caar elm) '((nil	. "o ")
					       (@bullet . "* ")
					       (@minus	. "- "))))))
       (if (not style)
	   (error "oops:@%s %s" op (buffer-substring-no-properties (point) e)))
       ;; delete param
       (delete-region (point) e)
       (and (looking-at "\n") (delete-char 1))
       (list :style style)))
    ((enumerate)
     (let* ((e (point-at-eol))
	    (elm (gauche-read-from-string
		  (buffer-substring-no-properties (point) e))))
       (delete-region (point) e)
       (if (and (null (cdr elm))
		(>= 1 (length (car elm))))
	   (cond
	    ((not (car elm))
	     '(:style "%d. " :start 1))
	    ((and (integerp (caar elm))
		  (or (zerop (caar elm)) (plusp (caar elm))))
	     (list :style "%d. " :start (caar elm)))
	    ((<= ?a (caar elm) ?z)
	     (list :style "%c. " :start (caar elm)))
	    ((<= ?A (caar elm) ?Z)
	     (list :style "%c. " :start (caar elm)))
	    (t
	     (error "oops:@%s %s elm:%S"
		    op
		    (buffer-substring-no-properties (point) e)
		    elm)))
	 (error "oops:@%s %s elm:%S"
		op
		(buffer-substring-no-properties (point) e)
		elm))))
    ((multitable)
     (let* ((e (point-at-eol))
	    (param (buffer-substring-no-properties (point) e))
	    (ret (cond
		  ((eq (read param) '@columnfractions)
		   (let ((width (frame-width)))
		     (mapcar (lambda (f)
			       (floor (* f width)))
			     (cadr (gauche-read-from-string param)))))
		  (t
		   (save-restriction
		     (narrow-to-region (point) e)
		     (save-excursion
		       (do ((acc))
			   ((>= (point) (point-max)) (reverse acc))
			 (skip-chars-forward " \t")
			 (let ((b (point)))
			   (forward-sexp)
			   (push (- (length (buffer-substring-no-properties
					     b (point)))
				    (length "{}")) ;for "{}" where @cmd{op}
				 acc)))))))))
       (delete-region (point) e)
       (list :colwidths ret)))
    ((ifset ifclear)
     ;;
     )
    ((table ftable vtable)
     (let* ((e      (point-at-eol))
	    (elm    (gauche-read-from-string
		     (buffer-substring-no-properties (point) e)))
	    (flface (and (null (cdr elm))
			 (= 1 (length (car elm)))
			 (assq-val (caar elm) '((@asis . default) ;xxx
						(@b    . bold)
						(@code . gauche-doc-code-face)
						(@emph . bold-italic)
						(@i    . italic)
						(@r    . default)
						(@var  . gauche-doc-var-face))))))
       (if (not flface)
	   (error "oops:@%s %s" op (buffer-substring-no-properties (point) e)))
       ;; delete param
       (delete-region (point) e)
       (and (looking-at "\n") (delete-char 1))
       (list :first-line-face flface)))
    ;;XXX
    ((group tex)
     )
    ((cartouche)
     )
    ((flushleft flushright)
     )
    ))

;;; shift right
(defun gauche-texi-set-left-margin (ext)
  (let ((beg (extent-start-position ext))
	(end (extent-end-position ext))
	(gly (make-glyph (vector 'string
				 :data
				 (make-string
				  (extent-property ext :indent)
				  ?\ )))))
    (set-glyph-face gly 'default)
    (save-excursion
      (set-buffer (extent-object ext))
      (save-restriction
	(narrow-to-region beg end)
	;;untabify-bol
	(goto-char beg)
	(while (re-search-forward "^ *\t[ \t]*" nil t)
	  (let ((pos (point))
		(col (current-column))
		(indent-tabs-mode nil))
	    (delete-region (match-beginning 0) pos)
	    (indent-to col)))
	;;
	(goto-char beg)
	(while (re-search-forward "^." nil t)
	  (gd-make-left-margin gly (match-beginning 0)))))))

;; XXX: labels?
;;sub defuns for gauche-documentation-1
(eval-when-compile
  (defmacro gd-update-entry-type (def)
    (let ((gdef (make-symbol "gdef")))
      `(let* ((,gdef ,def)
	      (type (car ,gdef)))
	 (setcar ,gdef (cond
			 ((assq-val type '((defun   . "Function")
					   (defspec . "Special Form")
					   (defvar  . "Variable")
					   (defmac  . "Macro")
					   (defivar . "Instance Variable"))))
			((stringp type)
			 (if (char= (string-to-char type) ?<)
			     type
			   (capitalize-string-as-title type)))
			(t
			 (format "??%S" type)))))))
  (defmacro gd-make-docs (entry-list)
    `(mapcar (lambda (ent)
	       (let ((docs (gauche-doc-database ent)))
		 (dolist (def (car docs))
		   (gd-update-entry-type def))
		 docs))
	     ,entry-list))
  (defmacro gd-make-type-list (docs)
    `(mapcar (lambda (defs)
	       (caar (remove-if-not (lambda (e)
				      (string-equal (cadr e) symname))
				    (car defs))))
	     ,docs)))

(defun gauche-documentation-1 (entry-list symname)
  (with-displaying-help-buffer
   (lambda ()
     (with-current-buffer standard-output
       (set-syntax-table gauche-texi-syntax-table)
       (let* ((docs (gd-make-docs entry-list))
	      (types (gd-make-type-list docs)))
	 ;; multiple definitions
	 (when (> (length types) 1)
	   ;; Symbol - Type, Type...
	   (insert-face (concat symname " - " (string-join types ", ")) 'bold)
	   (let ((plst (go-props (go-intern-soft symname))))
	     (if plst
		 (insert "\n\t"
			 (mapconcat (lambda (elm)
				      (format "%s(%s)" (cdr elm) (car elm)))
				    plst
				    ", "))))
	   (insert "\n----------------------------------------\n"))
	 ;;
	 (do ((docs docs (cdr docs)))
	     ((null docs))
	   (dolist (def (caar docs))
	     ;; insert entry type
	     (insert-face (car def) 'bold)
	     (insert ": ")
	     ;; insert function/entry
	     (let ((b (point))
		   (f (cadr def)))
	       (insert f)
	       (if (string-equal symname f)
		   (set-extent-face (make-extent b (point)) 'highlight))
	       (if (string-match "@" f) ; for @var{TAG}xyz
		   (gauche-doc-detexi-and-set-attr b (point))))
	     (goto-char (point-max))
	     ;; has args
	     (if (cddr def)
		 (let ((b (point)))
		   (insert " " (cddr def))
		   ;;add arg attribute
		   (with-syntax-table gauche-mode-syntax-table
		     (save-excursion
		       (goto-char b)
		       ;; xxx
		       (while (re-search-forward "\\(?:\\w\\|\\s_\\)+" nil t)
			 ;;skip &(optional|keyword|...)
			 (if (not (eq (char-after (match-beginning 0))
				      ?&))
			     (gauche-append-text-property (match-beginning 0)
							  (match-end 0)
							  'face
							  'gauche-doc-var-face)))))))
	     (goto-char (point-max))
	     (insert "\n"))
	   ;;Documentation body...
	   (save-restriction
	     (narrow-to-region (point) (point))
	     (insert "\n"		;need "\n"
		     (cdar docs)
		     "\n@end default")
	     (gauche-doc-detexi-and-set-attr (point-min) (point-max)
					     (list 'default))
	     (do ((l (let (acc)
		       (map-extents
			(lambda (e d)
			  (let ((beg (extent-start-position e))
				(end (extent-end-position e)))
			    (if (not (memq beg acc))
				(push beg acc))
			    (if (not (memq end acc))
				(push end acc)))
			  nil)
			nil (point-min) (point-max) nil nil 'gauche-texi-env)
		       (sort acc #'<)) (cdr l)))
		 ((null (cdr l)))
	       (let* ((prop (extent-property (extent-at (car l) nil 'gauche-texi-env)
					     'gauche-texi-env))
		      (indent (plist-get (cdr prop) :indent))
		      (ext (make-extent (car l) (cadr l))))
		 (set-extent-properties ext (list :fill (gd-texi-attr (car prop) :fill)
						  :indent indent))))
	     ;;fill-region
	     (map-extents (lambda (e d)
			    (let ((fill-column (- (frame-width)
						  (extent-property e :indent)
						  gauche-doc-right-margin)))
			      (fill-region (extent-start-position e)
					   (extent-end-position e)))
			    nil)
			  nil (point-min) (point-max) nil nil :fill t)
	     ;;add left-margin
	     (map-extents (lambda (e d)
			    (gauche-texi-set-left-margin e)
			    nil)	;nil for loop
			  nil (point-min) (point-max) nil nil :indent))
	   ;;
	   (when (cdr docs)
	     (goto-char (point-max))	;for "" doc.
	     (insert "\n---\n"))))))
   ;; Buffer name
   (format "Document for `%s'" symname)))

(defun gauche-doc-db-handle ()
  "Return the berkeley db file handler.
See also variable *gauche-doc-db*'s document."
  (cond
   ((database-live-p *gauche-doc-db*)
    *gauche-doc-db*)
   ((stringp gauche-doc-file-name)
    (let* ((file (concat (file-name-sans-extension gauche-doc-file-name)
			 ".db"))
	   (db (open-database file 'berkeley-db 'btree "r")))
      (if (not db)
	  (error "Cannot open document db file(%s)." file))
      (setq *gauche-doc-db* db)))
   (t
    (error "set gauche-doc-file-name"))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Inferior Gauche functions
;;
(eval-when-compile
  (defmacro gauche-set-send-strings (lst)
    `(progn (setq *gauche-send-strings* ,lst)
	    (gauche-update-modeline)))
  (defmacro gauche-read-from-stream (stream)
    `(save-excursion
       (set-buffer ,stream)
       (save-restriction
	 (widen)
	 (goto-char (point-max))
	 ;; Is there next prompt?
	 (when (and (re-search-backward "\^@" nil t)
		    (> (match-beginning 0) 1))
	   (goto-char 1)
	   (let ((expr (read (current-buffer)))
		 (sstr (car *gauche-send-strings*)))
	     ;;(skip-chars-forward " \t\n\0")
	     (skip-chars-forward " \t\n") ;left "\0"
	     (unwind-protect
		 (delete-region 1 (point))
	       (gauche-set-send-strings (cdr *gauche-send-strings*)))
	     (cons sstr expr))))))
  (defmacro go-send-string-to-gauche (process string)
    (let ((proc (make-symbol "proc")))
      `(let ((proc ,process))
	 (process-send-string proc ,string)
	 (process-send-string proc "\n")
	 (process-send-eof    proc)))))

(define-error 'gauche-mode "Gauche error")
(define-error 'gauche-not-running
  "gosh is not running" 'gauche-mode)
(define-error 'gauche-process-busy
  "gosh running, but not ready" 'gauche-mode)
(define-error 'gauche-process-connection-type
  "gauche-mode requires pty connection" 'gauche-mode)
(define-error 'gauche-completion-function
  "Invalid predicate for completion." 'gauche-mode)
;;; functions for inferior process
(defun gauche-process (&optional process-name)
  "Return the gauche process."
  (get-process (or process-name gauche-inferior-process-name)))

(defun gauche-inferior-process-live-p (&optional process)
  (process-live-p (or process
		      (gauche-process gauche-inferior-process-name))))

(defun gauche-stream-buffer (buf)
  (cond ((buffer-live-p buf)
	 buf)
	((stringp buf)
	 (or (get-buffer buf)
	     (setq *gauche-stream*
		   (set-process-buffer (gauche-process)
				       (get-buffer-create buf)))))
	(t
	 (setq *gauche-stream*
	       (set-process-buffer (gauche-process)
				   (get-buffer-create gauche-stream-name))))))

(defun gauche-cleanup-streams ()
  (unwind-protect
      (erase-buffer (gauche-stream-buffer *gauche-stream*))
    (gauche-set-send-strings nil)))

(defun gauche-start-inferior-repl ()
  "Start gauche(gosh) process, if gauche is not running.
You can add this function to gauche-mode-hook."
  (interactive)
  (if (not gauche-process-ptyp)
      (error 'gauche-process-connection-type
	     (format "Your connection type is %S." gauche-process-ptyp)))
  (unless (gauche-inferior-process-live-p)
    (let ((p (apply #'start-process
		    gauche-inferior-process-name
		    gauche-stream-name
		    gauche-program-name
		    (append gauche-program-args
			    '("-ugauche.interactive"
			      "-e(with-module igauche (repl))")))))
      (gauche-cleanup-streams)
      (process-kill-without-query p)
      (set-process-sentinel	  p 'gauche-inferior-process-sentinel)
      (set-process-filter	  p 'gauche-process-filter)
      (message "gauche process started."))))

(defun gauche-inferior-process-sentinel (process event)
  (case (process-status process)
    ((exit signal)
     (gauche-cleanup-streams)))
  (message "%s is %s" process event))

;; access to inferior process
(defun gauche-send-interrupt ()
  "Send interrupt signal to gauche process."
  (interactive)
  (if (not (gauche-inferior-process-live-p))
      (error 'gauche-not-running "gauche process is not running"))
  (unwind-protect
      (interrupt-process (gauche-process) gauche-process-ptyp)
    (gauche-cleanup-streams)))

(defun gauche-send-string-to-process (process string &optional sync)
  "Send PROCESS the contents of STRING as input. and get output message."
  (if (not (gauche-inferior-process-live-p))
      ;;support autostart?
      (error 'gauche-not-running
	     (substitute-command-keys
	      "Type ``\\[gauche-start-inferior-repl]'' to start gauche process")))
  (if sync
      (gauche-send-string-to-process-sync process string)
    (unwind-protect
	(gauche-set-send-strings (nconc *gauche-send-strings*
					(list (cons :in string))))
      (go-send-string-to-gauche process string))))

(defun gauche-send-string-to-process-sync (process string)
  (let ((n (length *gauche-send-strings*)))
    (if (> n 0)
	(error 'gauche-process-busy
	       (format "evaluating previous sexp%s(%s)."
		       (if (> n 1) "s" "")
		       n))))
  (let ((buf     (process-buffer process))
	(ofilter (process-filter process)))
    (with-temp-buffer
      (unwind-protect
	  (progn
	    (set-process-filter  process nil)
	    (set-process-buffer  process (current-buffer))
	    (go-send-string-to-gauche process string)
	    (gauche-update-modeline 'evaluating)
	    (accept-process-output process)
	    (while (not (char= (char-before (point-max)) ?\^@))
	      (accept-process-output process))
	    (if (input-pending-p)
		(discard-input)))
	(set-process-filter process ofilter)
	(set-process-buffer process buf)
	(gauche-update-modeline))
      (goto-char (point-min))
      (read (current-buffer)))))

;; from scheme.el
(defvar scheme-mode-line-process "")
;; modeline-update function
(defun gauche-update-modeline (&optional sym)
  "update gauche-mode process status(scheme-mode-line-process).
process-status is
repl:-- ... no sexp, repl is idle,
repl:nn ... waiting sexps are evaluated,
repl:sym if SYM is Non-NIL."
  (setq scheme-mode-line-process
	(cond ((not (gauche-inferior-process-live-p))
	       "")
	      (sym
	       (format " repl:%s" sym))
	      (t
	       (format " repl:%02.2s" (let ((n (length *gauche-send-strings*)))
					(if (zerop n)
					    "--"
					  n)))))))

(defun gauche-send-region-to-process (process start end &optional buffer sync)
  "Send current contents of the region between START and END as input to
 PROCESS. PROCESS may be a process or the name of a process, or a buffer
or the name of a buffer, in which case the buffer's process is used.
If it is nil, the current buffer's process is used.
 BUFFER specifies the buffer to look in; if nil, the current buffer is used."
  (gauche-send-string-to-process process
				 (buffer-substring-no-properties start end)
				 sync))

(defun gauche-eval-string-sync (string &optional module key noerror)
  "Send STRING to gauche process, gauche evaluates it.
This function returns the gauche's stdout as alist.
If MODULE is specified, STRING is evaluated (with-module MODULE STRING).
If KEY is specified, return only specified alist."

  (let ((alist (gauche-send-string-to-process (gauche-process)
					      (if module
						  (format "(with-module %s %s)"
							  module string)
						string)
					      'sync)))
    (if noerror
	alist
      (if (assq :err alist)
	  (error 'gauche-mode (assq-val :err alist)))
      (let ((out (assq-val :out alist)))
	(case key
	  ((:disp :vals) (remove-if-not (lambda (e)
					  (eq (car e) key))
					out))
	  ((nil)         out)
	  (t             (error "unknown key(%S)" key)))))))

(defun gauche-eval-region (start end)
  "Send the current region to the inferior gauche process.
And echo the output-string on echo area.
If CONT is non-nil, call (CONT output-string)."
  (interactive "r")
  (gauche-send-region-to-process (gauche-process) start end))

(defun gauche-eval-buffer (buffer)
  "Send buffer to the inferior gauche process."
  (interactive "bGauche eval whole buffer: ")
  (save-excursion
    (set-buffer buffer)
    (save-restriction
      (widen)
      (gauche-eval-region (point-min) (point-max)))))

(defun gauche-eval-definition ()
  "Send the current definition to the inferior gauche process."
  (interactive)
  (save-excursion
    (end-of-defun)
    (let ((end (point)))
      (beginning-of-defun)
      (gauche-eval-region (point) end))))

(defun gauche-eval-last-sexp ()
  "Send the previous sexp to the inferior gauche process."
  (interactive)
  (gauche-eval-region (save-excursion (backward-sexp) (point)) (point)))

(defun gauche-eval-last-sexp-and-print ()
  "Send the previous sexp to the inferior gauche process. And
insert its output."
  (interactive)
  (let ((alist (gauche-eval-string-sync
		(buffer-substring-no-properties
		 (save-excursion (backward-sexp) (point)) (point))
		nil nil t)))
    (insert "\n")
    (dolist (elm alist)
      (case (car elm)
	((:out)
	 (dolist (e (cdr elm))
	   (case (car e)
	     ((:vals)
	      (insert-face (cdr e) 'gauche-output-values-face)
	      (insert "\n"))
	     ((:disp)
	      (insert-face (cdr e) 'gauche-output-display-face)))))
	((:err)
	 (insert-face (cdr elm) 'gauche-output-error-face))))))

;; emacs 21.3.x's message supports multi-line echo
;; xemacs 21.4.x's message dose not, we need kludge.
(defun gauche-message (fmt &rest args)
  "multi-line echo. like a emacs 21.3.x's one."
  (let ((buf (get-buffer " *Echo Area*"))
	(win (minibuffer-window)))
    (save-excursion
      (set-buffer buf)
      (save-window-excursion
	(set-window-buffer win buf 'norecord)
	(apply #'lmessage 'no-log fmt args)
	(goto-char (point-max))
	(skip-chars-backward " \f\r\n\t")
	;; XXX
	(do ((i 1 (1+ i))
	     ;;;XXX
	     (m (* (frame-height) (min 0.35 temp-buffer-max-height))))
	    ((or (>= i m) (pos-visible-in-window-p nil win))
	     (and (> i 1) (or (input-pending-p)
			      (sit-for gauche-message-timeout))))
	  (enlarge-window 1 nil win))))))

(defun gauche-append-output (alist)
  "This funcation called by gauche-process-filter.
ALIST is ((:in  . send-string to gauche-process)
	  (:out (:disp . displayed-string) (:vals . value-string))
	  (:err . error-message))"
  (save-excursion
    (set-buffer (get-buffer-create gauche-inferior-output-buffer-name))
    (goto-char (point-max))
    (insert (format "%S\n" alist))))

(defun gauche-show-output (alist)
  "This funcation called by gauche-process-filter.
ALIST is ((:in  . send-string to gauche-process)
	  (:out (:disp . displayed-string) (:vals . value-string))
	  (:err . error-message))"
  (gauche-message "%s"
		  (if (not (cdr alist))	; no :err, :out
		      "\; no values returned." ; \ for emacs el parser
		    (concat (if (assq :out alist)
				(mapconcat (lambda (e)
					     (case (car e)
					       ((:disp)
						(cdr e))
					       ((:vals)
						(concat (cdr e) "\n"))
					       (t
						(error "unknown key(%S)"
						       (car e)))))
					   (assq-val :out alist)
					   ""))
			    (assq-val :err alist)))))
;;; Inferior gauche commands
(defun gauche-inferior-complete-symbol ()
  "Perform completion on gauche symbol preceding point.
Compare that symbol against the gosh known symbols.

The context determines which symbols are considered. If the symbol
starts just after an open-parenthesis, only syntaxes and functions.
In use form, only modules.  Otherwise, all symbols with function
definitions, values or properties are considered."
  (interactive)
  (gauche-complete-symbol-1 :function    (list #'gauche-inferior-completions
					       'syntax-or-procedure?)
			    :module      (list #'gauche-inferior-completions
					       'module)
			    :all-modules (list #'gauche-inferior-completions
					       'all-modules)
			    :default     (list #'gauche-inferior-completions
					       '\#t)))

(defun gauche-inferior-completions (string &optional predicate comps)
  "This Function is used for try-completion or all-completions.
Query to gosh and get result.
Return common substring of all completions of STRING in gauche symbols.
If optional second argument PREDICATE is non-nil, it is used to test
each prossible match. PREDICATE is evaluated by gosh. So, It must be
gosh form. For example, if you wish to get completion the procedures,
specify the `` 'procedure? ''.
If optional third argument comps is non-nil, function returns completions
list.
See try-completion/all-completions for more details."

  (read (cdar (gauche-eval-string-sync
	       (format "(igauche-completions %S '%s %s %s)"
		       string
		       (case predicate
			 ((module all-modules) predicate) ; collections
			 (t			 'symbol))
		       (case predicate
			 ((nil module all-modules) '\#f) ; predicate
			 (t                        predicate))
		       (if comps '\#t '\#f)) ;all-comps/try-comp
	       'igauche
	       :vals))))

;; from: hyper-apropos.el, apropos.el
(defun gauche-inferior-apropos (regexp)
  "Display lists of symbols matching REGEXP."
  (interactive (list (read-string "iGauche Apropos(regexp): ")))
  (let ((result (cdar (gauche-eval-string-sync
		       (format "(apropos #/%s/)" regexp) nil :disp))))
    (with-displaying-help-buffer
     (lambda ()
       (with-current-buffer standard-output
	 (if (not result)
	     (insert-face (format "Nothing for %s." regexp) 'bold)
	   (insert-face (format "Apropos search for: %s\n\n" regexp) 'bold)
	   (insert result))))
     (format "Apropos for `%s'" regexp))))

(defun gauche-inferior-describe-symbol (symbol)
  "Describe gauche symbol. When run interactively, it defaults to any symbol
found by `gauche-symbol-name-at-point'."
  (interactive
   (let ((sname (gauche-symbol-name-at-point)))
     (list (let ((enable-recursive-minibuffers t))
	     (completing-read (format "iGauche Symbol%s: "
				      (if (> (length  sname) 0)
					  (format " (default %s)" sname)
					""))
			      #'gauche-inferior-completions
			      nil
			      t nil nil sname)))))
  (if (string-equal symbol "")
      (message "Please enter symbol name.")
    (let ((result (cdar (gauche-eval-string-sync
			 (format "(describe %s)" symbol) nil :disp))))
      (with-displaying-help-buffer
       (lambda ()
	 (with-current-buffer standard-output
	   (insert-face (format "Describe symbol: %s\n\n" symbol) 'bold)
	   (insert result)))
       (format "Describe symbol `%s'" symbol)))))
;;
(defun gauche-view-module-source (module)
  "View module source file."
  (interactive
   (let ((enable-recursive-minibuffers t))
     (list (completing-read "iGauche module: "
			    #'gauche-inferior-completions
			    'module
			    t))))
  (let ((result (read (cdar (gauche-eval-string-sync
			     (format "(library-map '%s (lambda (s p) p))"
				     module)
			     nil
			     :vals)))))
    (if (= (length result) 1)
	(view-file-other-window (car result))
      (message "Got %S" result))))

;;xxx, need pretty-print, use slib?
(defun gauche-macroexpand-at-point ()
  (interactive)
  (gauche-macroexpand-1-at-point 'expand-all))

(defun gauche-macroexpand-1-at-point (&optional expand-all)
  (interactive)
  (let ((buf (format "*Gauche Macroexpand%s*" (if expand-all
						  ""
						"-1"))))
    (with-output-to-temp-buffer buf
      (princ (gauche-macroexpand (save-excursion
				   (skip-chars-forward " \t")
				   (let ((b (point)))
				     (forward-sexp 1)
				     (buffer-substring-no-properties b (point))))
				 expand-all))
      (set-buffer buf)
      (gauche-mode))))

(defun gauche-macroexpand (ssexp &optional expand-all)
  (cdar (gauche-eval-string-sync (format "(%s '%s)"
					 (if expand-all
					     'macroexpand
					   'macroexpand-1)
					 ssexp)
				 nil
				 :vals)))

;;+ from scheme.el
(defvar calculate-lisp-indent-last-sexp)

;; Copied from lisp-indent-function, but with gets of
;; scheme-indent-{function,hook}.
(defun gauche-indent-function (indent-point state)
  "indent function for gauche code.
This code copied from scheme.el. But function supports
``gauche-indent-regexp-alist''."
  (let ((normal-indent (current-column)))
    (goto-char (1+ (elt state 1)))
    (parse-partial-sexp (point) calculate-lisp-indent-last-sexp 0 t)
    (if (and (elt state 2)
             (not (looking-at "\\sw\\|\\s_")))
        ;; car of form doesn't seem to be a symbol
        (progn
          (if (not (> (save-excursion (forward-line 1) (point))
                      calculate-lisp-indent-last-sexp))
              (progn (goto-char calculate-lisp-indent-last-sexp)
                     (beginning-of-line)
                     (parse-partial-sexp (point)
					 calculate-lisp-indent-last-sexp 0 t)))
          ;; Indent under the list or under the first sexp on the same
          ;; line as calculate-lisp-indent-last-sexp.  Note that first
          ;; thing on that line has to be complete sexp since we are
          ;; inside the innermost containing sexp.
          (backward-prefix-chars)
          (current-column))
      (let* ((function (buffer-substring-no-properties (point)
						       (progn
							 (forward-sexp 1)
							 (point))))
	     (method (or (get (go-intern-soft function)
			      'gauche-indent-function)
			 (get (go-intern-soft function)
			      'gauche-indent-hook)
			 (cdr (assoc* function
				      gauche-indent-regexp-alist
				      :test (lambda (str regex)
					      (string-match regex str)))))))
	(cond ((eq method 'defun)
	       (lisp-indent-defform state indent-point))
	      ((integerp method)
	       (lisp-indent-specform method state
				     indent-point normal-indent))
	      ((functionp method)
	       (funcall method state indent-point normal-indent)))))))
;;-

;;;
(defun gauche-process-filter (proc str)
  (let ((buf (gauche-stream-buffer *gauche-stream*)))
    (with-current-buffer buf
      (save-restriction
	(widen)
	(goto-char (point-max))
	(insert str)))
    (do ((ss-ret (gauche-read-from-stream buf) (gauche-read-from-stream buf)))
	((not ss-ret))
      (dolist (func (append gauche-process-filter-functions
			    (if (<= (buffer-size buf) 1) ; prompt "\0"
				gauche-process-filter-functions-on-last)))
	(funcall func ss-ret)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; gauche-interaction

;;;XXX
(defun gauche-complete-sexp-p (beg end)
  (save-excursion
    ;; for '(....), try begining of input string
    (goto-char beg)
    (save-restriction
      (narrow-to-region beg end)
      ;; parse-sexp-ignore-comments should be ``t''
      (do ((b (scan-sexps (point) 1 nil t) (scan-sexps (point) 1 nil t)))
	  ((not b))
	(goto-char b)))
    (skip-chars-forward " \t\r\n")
    (= (point) end)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; args & keywords

(defun gauche-function-args (sym)
  "Return the list of (type . argstring) for go-interned SYM."
  (if (gauche-syntax-or-procedure-p sym)
      (let ((args (get sym :args 'not-set))) ;nil meens no arg
	(if (eq args 'not-set)
	    (put sym :args (gauche-function-args-1 sym))
	  args))))

(defun gauche-function-args-1 (sym)
  "go-interned SYM"
  (mapcar (lambda (al)
	    (cons (car al) (and (cddr al) (gauche-detexi-string (cddr al)))))
	  (apply #'append
		 (let ((name (symbol-name sym)))
		   (mapcar (lambda (ent)
			     (remove-if-not (lambda (l)
					      (string-equal name (cadr l)))
					    (car (gauche-doc-database ent))))
			   (gauche-doc-database name))))))

(defun gauche-detexi-string (str)
  (with-temp-buffer
    (insert str)
    (set-syntax-table gauche-texi-syntax-table)
    (gauche-doc-detexi-and-set-attr (point-min) (point-max))
    (buffer-substring-no-properties (point-min) (point-max))))

(defun gauche-function-keywords-list (name)
  "Return function's keywords list."
  (let ((sym (go-intern-soft name)))
    (if (gauche-syntax-or-procedure-p sym)
	(let ((keywords (get sym :keywords 'not-set)))
	  (if (eq keywords 'not-set)
	      (put sym :keywords (gauche-function-keywords-list-1 sym))
	    keywords)))))

(defun gauche-function-keywords-list-1 (sym)
  "go-interned SYM"
  (mapcar (lambda (argstr)
	    (mapcar (lambda (elm)
		      (let ((s (symbol-name (cond ((null elm))
						  ((atom elm)
						   elm)
						  ;;(keyword default)
						  ((consp elm)
						   (car elm))))))
			;; gauche's keyword name is key,
			;; but we use :key style string
			(go-intern (if (char= ?: (string-to-char s))
				       s
				     (concat ":" s)))))
		    ;; some keyword's doc has following style
		    ;; ``&keyword key1 (key2 default)''
		    ;; Using read to get key1/2, but #t/f ...
		    (read (concat "(" (replace-in-string
				       ;; #t/f -> \\#t/f
				       (replace-in-string argstr "#" "\\\\#")
				       ;; delete @dots{}
				       "\\s-+\\.\\.\\.\\s-*" "")
				  ")"))))
	  (remove-if #'null
		     (mapcar (lambda (arg)
			       (let ((str (cdr arg)))
				 (if (and (stringp str)
					  (string-match "&keyword\\s-+" str))
				     (substring str (match-end 0)))))
			     (gauche-function-args sym)))))

(defun gauche-keyword-completions (patern funcname &optional comps)
  "Function for try/all-completions. Return the keyword list for FUNCNAME.
FUNCNAME is gauche function for keywords completion. This function depends on
current point. Specified keyword will not returned"
  (apply (if comps
	     #'all-completions
	   #'try-completion)
	 patern
	 (let ((cur (save-excursion
		      (backward-sexp)
		      (point))))
	   ;; (list collection prodicate)
	   (list (mapcar (lambda (sym)
			   (list (symbol-name sym)))
			 (apply #'append
				(gauche-function-keywords-list funcname)))
		 ;; predicate
		 `(lambda (alst)
		    (not (member (car alst)
				 ;; making keyword string list
				 ',(save-excursion
				     (re-search-backward (concat "(" funcname)
							 nil t)
				     ;; all words, patern matches only :key
				     (split-string
				      (buffer-substring-no-properties (point) cur))))))))))

;;; auto display function args
(defun gauche-toggle-auto-display-function-args (&optional arg)
  "toggle auto-display current function args."
  (interactive)
  (cond ((not (itimer-live-p gauche-function-args-idle-timer))
	 (setq gauche-function-args-idle-timer
	       (run-with-idle-timer gauche-function-args-idle-delay
				    t
				    #'gauche-show-current-function-args)))
	((and (not arg) (itimer-live-p gauche-function-args-idle-timer))
	 (delete-itimer gauche-function-args-idle-timer))))

;; copy from font-lock.el, based on font-lock-unique
(eval-when-compile
  (defmacro gd-unique (list)
    (let ((glist (make-symbol "glist"))
	  (ghead (make-symbol "ghead")))
      `(let* ((,glist ,list)
	      (,ghead ,glist))
	 (while ,glist
	   (setq ,glist (setcdr ,glist (delete (car ,glist) (cdr ,glist)))))
	 ,ghead))))

(defun gauche-show-current-function-args ()
  "Show function args at current point."
  (and (eq major-mode 'gauche-mode)
       (not this-command)
       (eq last-command 'self-insert-command)
       (char= last-input-char ?\ )
       (let ((func (gauche-function-at-point)))
	 (if func
	     (lmessage 'no-log
		 (mapconcat
		  #'identity
		  (gd-unique (mapcar
			      (lambda (elt)
				(if (cdr elt)
				    (format "(%s %s)"
					    func
					    (replace-in-string (cdr elt)
							       "&keyword.*"
							       "&keyword ..."))
				  (format "(%s)" func)))
			      (gauche-function-args (go-intern-soft func))))
		  "/"))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun gauche-show-last-output (num)
  (interactive "p")
  (let ((buf (get-buffer gauche-inferior-output-buffer-name))
	(bufname "Gauche Last Output"))
    (if (not buf)
	(error "no buf"))
    (with-displaying-help-buffer
     (lambda ()
       (with-current-buffer standard-output
	 (let ((lst (save-excursion
		      (set-buffer buf)
		      (goto-char (point-max))
		      (backward-sexp num) ;return nil when begining of buf
		      (do ((i num (1- i))
			   (acc))
			  ((or (<= i 0)
			       (= (point) (point-max)))
			   (nreverse acc))
			(push (read buf) acc)
			(skip-chars-forward " \t\n\l")))))
	   (dolist (elm lst)
	     (let ((in  (assq-val :in  elm))
		   (err (assq-val :err elm))
		   (out (assq-val :out elm)))
	       (if (stringp in)
		   (let ((p (point)))
		     (insert ";;" (replace-in-string in "\n[\t ]*" " " ))
		     (set-extent-face (make-extent p (point)) 'italic)
		     (insert "\n")))
	       (if out
		   (dolist (e out)
		       (cond ((eq (car e) :disp)
			      (insert-face (cdr e) 'gauche-output-display-face))
			     ((eq (car e) :vals)
			      (insert-face (cdr e) 'gauche-output-values-face)
			      (insert "\n"))
			     (t
			      (error "???")))))
	       (if (stringp err)
		   (insert-face err 'gauche-output-error-face)))))))
     bufname)))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; kill & cleanup

(defun gauche-kill-gauche-process ()
  "Send ``(exit)'' to gosh."
  (interactive)
  (gauche-send-string-to-process (gauche-process) "(exit)"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; xxx: donot use hash-table...
(defun gauche-make-doc-db (data-file)
  "gauche-doc.dat -> gauche-doc.db"
  (let ((h (make-hash-table :test #'equal :size 3000 :rehash-size 1.5))
	(gauche-document-copyright
	 (let ((key "gauche doc-db copyright"))
	   (cons key
		 (format "(((%S %S)) . %S)"
			 "copyright"
			 key
			 (concat 
			  ;; Gauche document's copyright(gauche-refe.texi)
			  "Copyright (c) 2001-2005 Shiro Kawai"
			  " (shiro@@acm.org).\n"
			  "See COPYING.Gauche(in BSD dir)"
			  " or top of "
			  data-file
			  ".\n"))))))
    ;; elm -- (((def sym . arg) (def sym . arg)) . doc-string)
    (mapcar (lambda (elm)
	      (let ((doc-key (concat "doc\0" (md5 (format "%S" elm)))))
		;;put doc\0md5has - doc
		(puthash doc-key elm h)
		;; def -- (def sym . arg)
		(mapcar (lambda (def)
			  (let* ((sym (cadr def))
				 (cur-key-list (gethash sym h))) ; for ``let''
			    (if (not (member doc-key cur-key-list))
				(puthash sym (cons doc-key cur-key-list) h))))
			;; return ((def sym . arg) (def sym . arg))
			;; e    -- (def sym . arg) 
			(dolist (e (car elm) (car elm))
			  ;; detexi arg-string
			  (if (and (cddr e) (string-match "@" (cddr e)))
			      (setcdr (cdr e) (gauche-detexi-string (cddr e))))))))
	    ;;((((def sym . arg) (def sym . arg)) . doc-string)
	    ;; (((def sym . arg)) . doc-string))
	    (read (with-temp-buffer
		    (insert-file-contents data-file)
		    (buffer-substring (point-min) (point-max)))))
    ;;
    (let* ((temp-db-name (make-temp-name (concat data-file ",,temp..")))
	   (db           (open-database temp-db-name 'berkeley-db 'btree 
					"rw+" #o644))
	   (db-file-name (concat (file-name-sans-extension data-file) ".db")))
      (if (not db)
	  (error "Oops"))
      (maphash (lambda (k v)
		 (let ((key (if (string-match "@" k)
				(gauche-detexi-string k)
			      k)))
		   (put-database key (format "%S" v) db)))
	       h)
      (let ((dockey (concat "doc\0" (md5 (cdr gauche-document-copyright)))))
	(put-database (car gauche-document-copyright)
		      (format "(%S)" dockey) db)
	(put-database dockey (cdr gauche-document-copyright) db))
      (close-database db)
      (if (file-exists-p db-file-name)
	  (copy-file db-file-name (make-backup-file-name db-file-name)
		     'overwrite 'keep-time))
      (rename-file temp-db-name db-file-name 'overwrite)
      (message "Wrote %s." db-file-name))))

(provide 'gauche-mode)

;;; Local Variables:
;;; eval: (put 'with-progress-message 'lisp-indent-function 1)
;;; End:
