;;
;;; igauche.scm - Inferior gauche for gauche-mode.
;;;
;;;   Copyright (c) 2004 - 2007 NISHIO Yasuhiro, All rights reserved.
;;;
;;;   Redistribution and use in source and binary forms, with or without
;;;   modification, are permitted provided that the following conditions
;;;   are met:
;;;
;;;   1. Redistributions of source code must retain the above copyright
;;;      notice, this list of conditions and the following disclaimer.
;;;
;;;   2. Redistributions in binary form must reproduce the above copyright
;;;      notice, this list of conditions and the following disclaimer in the
;;;      documentation and/or other materials provided with the distribution.
;;;
;;;   3. Neither the name of the authors nor the names of its contributors
;;;      may be used to endorse or promote products derived from this
;;;      software without specific prior written permission.
;;;
;;;   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;;;   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;;;   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;;;   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;;;   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;;   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
;;;   TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
;;;   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
;;;   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
;;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;;
;;;  $Id: igauche.scm,v 2.11 2007/09/03 17:32:36 nishio Exp $
;;;

(define-module igauche
  (use srfi-1)
  (use gauche.parameter)
  (export igauche-version)
  )
(select-module igauche)

(define-constant igauche-version
  "$Id: igauche.scm,v 2.11 2007/09/03 17:32:36 nishio Exp $")

;; util proedure for completion
(define (syntax-or-procedure? s)
  (or (procedure? s) (is-a? s <syntax>) (is-a? s <macro>)))

;;get & memorize module names
(define module-name-list
  (let ((mlist #f))
    (lambda options
      (let-keywords* options ((force-init? (not mlist))
			      (loadable-only? #f))
	(if force-init?
	    (set! mlist
		  (let loop ((s "*")
			     (l '()))
		    (let ((mods (library-map (string->symbol s)
				  (lambda (m p) m))))
		      (if (null? mods)
			  l
			  (loop (string-append s ".*") (append mods l)))))))
	(if (not loadable-only?)
	    (fold (lambda (s knil)
		    (if (member s knil)
			knil
			(cons s knil)))
		  mlist
		  (map module-name (all-modules)))
	    mlist)))))

(define (igauche-completions str collection predicate comps)
  (let ((len (string-length str)))
    ;; for try-completion
    (define (kons-try-comps ss cs)
      (if (null? cs)
	  (if (string=? str ss)
	      't			;T for xemacs
	      ss)
	  (do ((a (string->list (if (eq? cs 't) str cs)) (cdr a))
	       (b (string->list ss)			 (cdr b))
	       (i 0 (+ i 1)))
	      ((or (null? a) (null? b) (not (char=? (car a) (car b))))
	       (substring ss 0 i)))))
    ;; for all-completions
    (define (kons-all-comps ss knil)
      (if (member ss knil)
	  knil
	  (cons ss knil)))
    (define (kons-if-match s knil)
      (let ((ss (symbol->string s)))
	(if (or (< (string-length ss) len)
		(not (string=? str (substring ss 0 len)))
		(and (procedure? predicate)
		     (not (predicate (global-variable-ref 'user s)))))
;;		     (not (predicate (eval s (interaction-environment))))))
	    knil
	    ((if comps
		 kons-all-comps
		 kons-try-comps) ss knil))))
    (define (symbol-fold)
      (fold (lambda (mod knil)
	      (let ((export-syms (module-exports mod)))
		(if  (or (null? export-syms)
			 (eq? export-syms #t))
		     (hash-table-fold (module-table mod)
				      (lambda (s v knil)
					(kons-if-match s knil))
				      knil)
		     (fold kons-if-match knil export-syms))))
	    '()
	    (let ((m (find-module 'user)))
	      (append (module-imports m)
		      (module-precedence-list m)))))
    (define (module-fold)
      (fold kons-if-match '() (module-name-list :loadable-only? #t)))

    (define (all-modules-fold)
      (fold kons-if-match '() (module-name-list :loadable-only? #f)))

    (case collection
      ((module)
       (module-fold))
      ((all-modules)
       (all-modules-fold))
      (else
       (symbol-fold)))))

;(define (describe-symbol sym)
;  (if (symbol-bound? sym)
;      (with-module user (describe sym))
;      (format #t "``~s'' dose not have binding.\n" sym))
;  (values))

(define (igauche-initialize)
  (module-name-list :force-init? #t)
  (values))

(define-macro (print-and-init-stream stream)
  `(let ((str (get-output-string (,stream (open-output-string)))))
    (if (not (string=? str ""))
	(format #t "(:~a . ~s)" ',stream str))))

(define (repl)
  (let ((input  (delay (let ((str (port->string (current-input-port))))
			 (if (string=? str "")
			     (exit 0)
			     (open-input-string str)))))
	(output (open-output-string))
	(errout (open-output-string))
	(disp   (make-parameter (open-output-string)))
	(vals   (make-parameter (open-output-string))))
    ;;we need error-hander, to continue our repl.
    (let/cc break-inner-read-eval-print-loop
      (define (reader)
	(with-error-handler (lambda (e)
			      (report-error e)
			      (break-inner-read-eval-print-loop))
	  (lambda () (read (force input)))))
      (define (evaluator expr env)
	;; for expr bad-expr expr
	(with-error-handler (lambda (e)
			      (with-error-to-port errout
				(lambda ()
				  (report-error e)
				  (break-inner-read-eval-print-loop))))
	  (lambda ()
	    ;; separate error?
	    (with-error-to-port (disp)
	      (lambda ()
		(with-output-to-port (disp)
		  (lambda ()
		    (eval expr env))))))))
      (define (printer . args)
	(display (string-join (map (lambda (r)
				     (format #f "~s" r))
				   args)
			      ", ")
		 (vals)))
      (define (prompter)
	(print-and-init-stream disp)
	(print-and-init-stream vals))
      ;;
      (define (sigint-handler sig)
	(let* ((name (sys-signal-name sig))
	       (msg (format #f "unhandled signal ~a (~a)" sig name)))
	  (report-error (make <error> :message msg))
	  (break-inner-read-eval-print-loop)))

      (set-signal-handler! |SIGINT| sigint-handler)
      ;; prompt to emacs
      (display #\null)
      (flush)
      (let ((cout (make-parameter #f))
	    (cerr (make-parameter #f)))
	(dynamic-wind
	    (lambda ()
	      (cout (get-output-string output))
	      (cerr (get-output-string errout)))
	    (lambda ()
	      (with-error-to-port errout
		(lambda ()
		  (with-output-to-port output
		    (lambda ()
		      (read-eval-print-loop reader evaluator printer prompter))))))
	    (lambda ()
	      (let ((out (string-scan (get-output-string output) (cout)
				      'after))
		    (err (string-scan (get-output-string errout) (cerr)
				      'after)))
		(if (or (not (string=? out ""))
			(not (string=? err "")))
		    (format #t "(~a~a)\n"
			    (if (string=? out "")
				""
				(format #f "(:out ~a)" out))
			    (if (string=? err "")
				""
				(format #f "(:err . ~s)" err))))))))))
  (repl))

(provide "igauche")
