#!/usr/local/bin/gosh
;;;
;;; make-doc.scm - make doc s-expr for gauche-mode.
;;;  
;;;   Copyright (c) 2004, 2005 NISHIO Yasuhiro, All rights reserved.
;;;   
;;;   Redistribution and use in source and binary forms, with or without
;;;   modification, are permitted provided that the following conditions
;;;   are met:
;;;   
;;;   1. Redistributions of source code must retain the above copyright
;;;      notice, this list of conditions and the following disclaimer.
;;;  
;;;   2. Redistributions in binary form must reproduce the above copyright
;;;      notice, this list of conditions and the following disclaimer in the
;;;      documentation and/or other materials provided with the distribution.
;;;  
;;;   3. Neither the name of the authors nor the names of its contributors
;;;      may be used to endorse or promote products derived from this
;;;      software without specific prior written permission.
;;;  
;;;   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;;;   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;;;   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;;;   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;;;   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;;   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
;;;   TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
;;;   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
;;;   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
;;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;;  
;;;  $Id: make-doc.scm,v 1.18 2007/04/12 17:17:30 nishio Exp $
;;;

(use srfi-13)
;; @def(???) list
(define texi-def '("fn" "ivar" "mac" "spec" "tp" "un" "var" "vr"))

;; get (("@def??" ...) ("@def??" ...) ...)
(define (doc-list port beg-matcher end-matcher)
  ;; get ("@def??" "@def??x" "line1" "line2"
  (define (get-next)
    (let loop ((line (read-line port)))
      (cond ((eof-object? line) line)
	    ((beg-matcher line) => (lambda (m)
				     (get-def (m 1) (list line))))
	    (else (loop (read-line port))))))

  (define (get-def kind def)
    (define (findex->def func)
      (let* ((defs (string-append "(def(?:"
				  (string-join texi-def "|") "))"))
	     (matcher (string->regexp (string-append "^(@"
						     defs
						     "x?(?:\\s+{(.*?)})?)\\s+\\S+"))))
	(cond ((matcher (car def)) => (lambda (m)
					(string-append (m 1) " " func (m 'after)))))))
    ;;
    (let loop ((line (read-line port))
	       (def def))
      (cond ((eof-object? line) (error "lost closing"))
	    ((end-matcher line) => (lambda (m)
				     (if (string=? (m 1) kind)
					 (reverse def)
					 (loop (read-line port)
					       (cons line def)))))
	    ;; function index
	    ((#/^\s*@findex\s+(\S+)/ line) => (lambda (m)
						(loop (read-line port)
						      (cons (findex->def (m 1)) def))))
	    ;omit "@?index key
	    ((#/^\s*@\w+index\s*\S+\s*$/ line)
	     (loop (read-line port) def))
	    (else
	     (loop (read-line port) (cons line def))))))
  (let loop ((def (get-next))
	     (docs '()))
    (if (eof-object? def)
	docs
	(loop (get-next) (cons def docs)))))

;;
;; ("@def??" "@def??x" "line1" "line2" ...)
;;	 => (((def?? func args..) (@def??x func args..)) . "body")
(define (defs-body lst)
  ;; abc@code{symbol}xyz => abcsymbolxyz, cannot hanlde @code{ss{}ss}
  (define (de-at-code string)
    (cond ((#/@code{([^{}]+)}/ string)
	   => (lambda (m)
		(string-append (m 'before 0) (m 1) (de-at-code (m 'after 0)))))
	  (else string)))
  ;;H:
  (define (strip-br str)
    (cond ((#/^{(.+)}$/ str) => (lambda (m) (m 1)))
	  (else str)))
  ;
  (define (string->readable str)
    (open-sqb->opt (dot->dot-sexp str)))

  ;; "func args" => ("func" . "args")
  (define (line->symarg string)
    (cond ((#/^\s*({[^{}]+}|\S+)\s*(.+)?$/ string)
	   => (lambda (m)
		(let ((sym (strip-br (de-at-code (m 1)))) ; get symbol name
		      (arg (or (m 2) '())))
		  (cons sym arg))))))
  (let* ((defs (string-append "(def(?:" (string-join texi-def "|") "))"))
	 (matcher (string->regexp (string-append "^@"
						 defs
						 "x?(?:\\s+{(.*?)})?\\s+"))))
	   (let loop ((lst lst)
		      (def-list '())	; ((defun sym . arg-str) ...)
		      (deftype #f)	; nest check (deftp defivar eivar etp)
		      (body '()))
	     (cond ((null? lst)
		    (cons (reverse def-list)
			  (string-join (reverse body) "\n")))
		   ((matcher (car lst))
		    => (lambda (m)
			 (let ((def (string->symbol (m 1)))
			       (sub (and (m 2) (string-downcase (m 2))))
			       (synop (m 'after 0)))
			   (loop (cdr lst) 
				 (cons (cons (or sub def)
					     (if (equal? "reader syntax" sub)
						 (list synop)
						 (line->symarg synop)))
				       def-list)
				 (or deftype def)
				 
				 (if (or (equal? deftype def) (not deftype))
				     body
				     (cons (car lst) body))))))
		   (else
		    (loop (cdr lst) def-list deftype (cons (car lst) body)))))))

;;;
(define (make-doc port)
  (define (process lst)
    (for-each display
	      (list ";;\n"
		    ";; This file was converted from gauche-ref*.texi "
		    "by make-doc.scm.\t"
		    (sys-asctime (sys-localtime (current-time)))
		    ";; gauche-ref*.texi has following copyright notice.\n"
		    ";;\n;; Copyright (c) 2001-2005 Shiro Kawai (shiro@acm.org)\n"
		    ";;\n"
		    ";; This data(document) is covered above copyright.\n"
		    ";;\n"))
    (display "(")
    (for-each
     (lambda (l)
       (let* ((elm  (defs-body l))
	      (defs (car elm))
	      (doc  (cdr elm))
	      (type (caar defs))
	      (skip (and (string? type)
			 (#/^program|category|command option$/ type))))
	 (if (not skip)
	     (begin (write elm) (newline)))))
     lst)
    (display ")\n"))
  ;;
  (let ((defs (string-append "(def(?:" (string-join texi-def "|") "))")))
    (process (doc-list port
		       (string->regexp (string-append "^@"
						      defs 
						      "\\s"))
		       (string->regexp (string-append "^@end\\s+"
						      defs
						      "\\s*$"))))))
;;;
(define (main args)
  (if (null? (cdr args))
      (make-doc (current-input-port))
      (for-each (lambda (f)
		  (call-with-input-file f (lambda (p) (make-doc p)) :encoding 'utf-8))
		(cdr args)))
  (exit 0))
