#!/usr/local/bin/gosh
;;;
;;; make-sym.scm - make gauche symbols s-expr for gauche-mode.
;;;  
;;;   Copyright (c) 2004-2006 NISHIO Yasuhiro, All rights reserved.
;;;   
;;;   Redistribution and use in source and binary forms, with or without
;;;   modification, are permitted provided that the following conditions
;;;   are met:
;;;   
;;;   1. Redistributions of source code must retain the above copyright
;;;      notice, this list of conditions and the following disclaimer.
;;;  
;;;   2. Redistributions in binary form must reproduce the above copyright
;;;      notice, this list of conditions and the following disclaimer in the
;;;      documentation and/or other materials provided with the distribution.
;;;  
;;;   3. Neither the name of the authors nor the names of its contributors
;;;      may be used to endorse or promote products derived from this
;;;      software without specific prior written permission.
;;;  
;;;   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;;;   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;;;   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;;;   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;;;   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;;   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
;;;   TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
;;;   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
;;;   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
;;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;;  
;;;  $Id: make-sym.scm,v 1.29 2007/04/12 16:54:41 nishio Exp $
;;;

(use file.util)
(use srfi-1)				;for (member item list eq-proc)
(use srfi-13)

;; make-sym.scm exclude-mod...
;;
;; Don't taint user module
(define (main argv)
  ;; print file header(comments).
  (define (print-comments)
    (for-each display
	      (list ";;\n;; This file was made by make-sym.scm.\n"
		    ";; Do NOT edit this file.\n"
		    (format #f ";; exclude modules:\t~s\n" (cdr argv))
		    (format #f ";; build at:\t~a"
			    (sys-asctime (sys-localtime (current-time))))
		    ";;\n")))
  ;; make gauche.ioutil rfc.???...
  (define (module-name-list s l)
    (let ((mods (library-fold (string->symbol s)
			      (lambda (m p knil)
				;; srfi-1's member has eq-proc
				(if (member (symbol->string m)
					    (cdr argv)
					    (lambda (i l)
					      (equal? (string-scan i l) 0)))
				    knil
				    (cons m knil)))
			      '())))
      (cond ((null? mods) l)
	    (else
	     (module-name-list (string-append s ".*") (append mods l))))))
  ;; make (module-name (<class1> . symbol1) (<class2> . symbol2)...)
  (define (module-and-class-symbol-pair mod)
    (cons (module-name mod)
	  (let ((exports-sym (module-exports mod)))
	    (cond ((or (eq? #t exports-sym)
		       (and (null? exports-sym)
			    ;; for default modules 
			    (memq mod
				  (module-precedence-list (current-module)))))
		   (hash-table-map (module-table mod)
				   (lambda (s v) s)))
		  (else
		   exports-sym)))))
  ;;
  (define (print-module-symbols mod-syms)
    (let ((mod  (car mod-syms))
	  (syms (cdr mod-syms)))
      (define (print-class-symbol-pair s)
	(with-error-handler (lambda (e) e)
	  (lambda ()
	    (let ((value (let ((v (global-variable-ref mod s)))
			   (if (is-a? v <autoload>)
			       (eval `(with-module ,mod ,s) 
				     (interaction-environment))
			       v))))
	      (format #t "\n  (~s . ~s)"
		      (symbol->string (class-name (class-of value)))
		      ;; |a b| -> "a b" for emacs
		      (symbol->string s))))))
      (format #t "(~s" (symbol->string mod))
      (for-each print-class-symbol-pair
		syms)
      (display ")\n ")))
  ;; load all mofules
  (for-each (lambda (mod)
	      (with-error-handler (lambda (e) #f)
		(lambda ()
		  (let ((f (module-name->path mod)))
		    (if (not (provided? f))
			(load f))))))
	    (module-name-list "*" '()))
  ;; print data.
  (print-comments)
  (display "(")
  (for-each print-module-symbols
	    (map module-and-class-symbol-pair (all-modules)))
  (display ")\n")
  (exit 0))
